/* Expose
 * Copyright (C) 1999 - Jeff Heflin - University of Maryland
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Author:  Jeff Heflin
 * Contact: Jeff Heflin (heflin@cs.umd.edu)
 *
 */

/*change history:
5-30-02 ZPan - change getInterfaceForEditKb() and getInterfaceForNewKb()
to abstrct, they will be implemented by ShoeParka.ExposeAPP
*/

package Expose;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Date;              // don't import java.util.List
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.TimeZone;
import java.util.Vector;
import java.net.*;
import java.text.*;                // for formatting and parsing dates
import SHOE.*;
import ShoeKb.*;
import HTML.*;
import UserInt.Utilities;

/** A web crawler that searches for SHOE web pages and stores the
  knowledge contained within them in a Parka knowledge base. A
  GUI allows the user to adjust parameters and monitor the progess of
  the crawler. */
/* This class is written for Java 1.1 since it is not intended to be run
   from an applet in a Web page */
public abstract class ExposeApp extends Frame
     implements ActionListener,
                Runnable
{
  /* Defaults for user-configurable properties */
  protected static String DEFAULT_FROM_USER  = "you@hostname";
  protected static String DEFAULT_KB_HOST    = "";
  protected static int    DEFAULT_KB_PORT    = 0;
  protected static String DEFAULT_KB_NAME    = "";
  protected static int    DEFAULT_MAX_PAGES  = 300;
  protected static int    DEFAULT_MAX_COST   = 20;
  protected static int    DEFAULT_REQUEST_INTERVAL = 30000;
  protected static boolean DEFAULT_IMPLICIT_CLAIMS = true;
  protected static boolean DEFAULT_USE_PROXY = false;

  public static final int    TABSIZE = 8;

  /* Page Costing constants */
  public static final int C_SHOE_PAGE = 1;
  public static final int C_HTML_PAGE = 5;
  public static final int C_DIFF_DIR  = 2;

  /** Name of the webcrawler */
  protected String userAgent = "Expose_SHOE_Robot";

  /* User-configurable properties */
  /** E-mail of the user who is running the agent. */
  protected String fromUser     = DEFAULT_FROM_USER;
  /** Hostname of machine on which to store the KB */
  protected String kbHost       = DEFAULT_KB_HOST;
  /** Port number for accessing KB on host machine */
  protected int    kbPort       = DEFAULT_KB_PORT;
  /** Name of KB to store data in */
  protected String kbName       = DEFAULT_KB_NAME;
  /** URL from which to begin the search. */
  protected String startUrl     = "";
  /** Vector of prefixes that sites must have to be visited. */
  protected Vector allowedUrls;
  /** Vector of prefixes for paths to avoid. */
  protected Vector prohibUrls;
  /** Limits the total number of pages that will be searched. */
  protected int    maxPages     = DEFAULT_MAX_PAGES;
  /** Limits the depth of the search. */
  protected int    maxCost      = DEFAULT_MAX_COST;
  /** The number of milliseconds between page requests to the same
    server. Polite agents wait 30 seconds. */
  protected int    requestInterval = DEFAULT_REQUEST_INTERVAL;
  /** Set to true if the use of an instance in a typed argument means
    that an additional claim should be made that the instance is of the
    correct type. */
  protected boolean implicitClaims = DEFAULT_IMPLICIT_CLAIMS;
  /** Set to true if the crawler should use a proxy server to
    access the Web. */
  protected boolean useProxy       = DEFAULT_USE_PROXY;
  /** A string of the form "userid:password" used to login to the
    proxy server */
  protected String  proxyLogin     = "";
  protected String  proxyHost      = "";
  protected String  proxyPort      = "";

  /* Button names */
  private static final String SEARCH  = " New KB  ";
  private static final String REVISIT = "Update KB";
  private static final String STOP    = "  Stop   ";
  private static final String RESUME  = " Resume  ";
  private static final String EXIT    = "  Exit   ";

  /* Status codes */
  /*
  public static final int STAT_URL_ERR = 1;
  public static final int STAT_EXCLUDED = 2;
  public static final int STAT_NOT_MOD = 3;
  public static final int STAT_SHOE = 4;
  public static final int STAT_SHOE_ERR = 5;
  public static final int STAT_NOT_SHOE = 6;
  */


  SHOE_Doc         doc;
  OntManager       ontman;                     // used to cache ontologies
  Exclusions       robotEx;
  KBInterface      kb = null;           // stores interface to knowledge base
  ShoeKb           shoeKb = null;
  Thread           searchThread;
  /** Encoded string that contains login information for the proxy server. */
  String           proxyAuth;

  /* search specfic variables */
  /** URLs to search */
  SearchList openList;
  /** URLs already considered */
  Hashtable closedList;
  /** URLs visited (doesn't include errors, exclusions, etc.). This is
    used to determine if a page has changed since last access and is
    used as the list of sites to visit when updating the KB. */
  Hashtable visitedList;


  /* counts */
  int searchCount = 0;
  int shoeCount = 0;
  int errCount = 0;

  DateFormat       dformat;
  SimpleDateFormat httpDate;
  String logFile = "expose.log";
  String initFile = "init.dat";
  String visitFile = "visited.dat";
  PrintStream logOut;

  Panel panelMain;
  Label labelStatus;
  TextField textURL;
  TextField textKB;
  List allowList;
  List avoidList;
  TextField textNewCon;
  TextArea detailText;
  TextField totCntText;
  TextField shoeCntText;
  TextField errCntText;

  Button buttonSearch;
  Button buttonRevisit;
  Button buttonStop;
  Button buttonResume;
  Button buttonExit;
  Button bAddAllow;
  Button bDelAllow;
  Button bAddAvoid;
  Button bDelAvoid;

  protected static final String PROXY_FILE = "ontproxy.dat";

  /*public static void main(String args[])
  {
    ExposeApp app;

    app = new ExposeApp();
  }*/

  /** Constructs a new Expose web-crawler. If the initialization and
   visited list files are present, it reads them and set the
   appropriate variables. */
  public ExposeApp()
  {
    super("Expose");

    /* initialize application variables */
    ontman = new OntManager();
    loadOntProxyFile();
    robotEx = new Exclusions();
    visitedList = new Hashtable();

    /* initialize user-configurable properties */
    allowedUrls = new Vector();
    prohibUrls = new Vector();

    /* date formatters (needed for visited file) */
    dformat = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.MEDIUM);
    dformat.setTimeZone(TimeZone.getDefault());
    httpDate = new SimpleDateFormat("EE, dd MMM yyyy HH:mm:ss z");
    TimeZone tz = TimeZone.getTimeZone("GMT");
    httpDate.setTimeZone(tz);
    // System.out.println("Date is " + httpDate.format(new Date()));

    /* File IO */
    try {
      logOut = new PrintStream(new FileOutputStream(logFile));
    } catch (IOException ex) {
      setStatus("Couldn't open log file " + logFile);
      return;
    }
    readInitFile();
    readVisitedFile();

    if (fromUser.equals("you@hostname")) {
      System.out.println("ERROR - You must set the FROM_USER property to " +
			 "your e-mail address.");
      System.out.println("\tYou can set this in the init.dat file or " +
			 "change the default value");
      System.out.println("\tin the source code.");
      exit();
    }

    if (useProxy == true)
      initProxyServer(proxyLogin, proxyHost, proxyPort);

    // the setup of the screen requires some of variables set above
    setup();
    pack();
    setSize(getMinimumSize());
    show();

  }



  /* ********************* Public methods *********************** */


  /** Implements the ActionListener interface to respond to button
    presses. */
  public void actionPerformed(ActionEvent event) {

    // String command = event.getActionCommand();
    Object source = event.getSource();

    /* start a new search from URL entered in text box */
    if (source == buttonSearch) {
      buttonSearch.setEnabled(false);
      buttonRevisit.setEnabled(false);
      buttonStop.setEnabled(true);
      buttonResume.setEnabled(false);

      /* ignore if search is already active */
      if(searchThread == null) {
	initializeSearch();               // reset search variables

	/* create KB */
	kb = getInterfaceForNewKb();      // prepare to create new KB file
	shoeKb = getShoeKb(kb, ontman, logOut);
	// shoeKb = new ShoeKb(kb, ontman, logOut);
	try {
	  shoeKb.establishConnection(kbHost, kbPort);
	  shoeKb.createKB(kbName);

	  visitedList = new Hashtable();    // start a new visited list
	  detailText.setText("");           // clear detail window
	  setStatus("searching...");
	  searchThread = new Thread(this);  // launch a thread to do the search
	  searchThread.start();
	} catch (KBException ex) {
	  System.out.println(ex.toString());
	}
      }
    }
    /* pause the search */
    else if (source == buttonStop) {
      buttonSearch.setEnabled(true);
      buttonRevisit.setEnabled(true);
      buttonStop.setEnabled(false);
      buttonResume.setEnabled(true);

      Thread temp = searchThread;
      searchThread = null;
      temp.interrupt();                   // wake up sleeping threads
      try {
	temp.join();                        // wait for thread to finish pass
      } catch (InterruptedException ex) {
	// VM wants us to continue
      }
      setStatus("Search stopped");

      /* should we close the KB? */
    }
    /* resume search using existing open and closed lists */
    else if (source == buttonResume) {
      buttonSearch.setEnabled(false);
      buttonRevisit.setEnabled(false);
      buttonStop.setEnabled(true);
      buttonResume.setEnabled(false);

      if (openList == null)
	return;
      /* ignore if search is already active */
      if(searchThread == null){
	setStatus("searching...");
	searchThread = new Thread(this);   // launch a thread to do the search
	searchThread.start();
      }
    }
    else if (source == buttonRevisit) {
      buttonSearch.setEnabled(false);
      buttonRevisit.setEnabled(false);
      buttonStop.setEnabled(true);
      buttonResume.setEnabled(false);

      VisitNode vnode;

      if (visitedList.isEmpty())
	return;
      /* ignore if search is already active */
      if(searchThread == null){
	initializeSearch();

	/* open KB */
	kb = getInterfaceForEditKb();
	shoeKb = getShoeKb(kb, ontman, logOut);
	// shoeKb = new ShoeKb(kb, ontman, logOut);
	try {
	  shoeKb.establishConnection(kbHost, kbPort);
	  shoeKb.openKB(kbName);
	} catch (KBException ex) {
	  System.out.println(ex.toString());
	  setStatus(ex.toString());
	}

	/* remove element added by initialize search, add elements from
	   the visited list */
	openList.removeElementAt(0);
	Enumeration enum = visitedList.elements();
	while (enum.hasMoreElements()) {
	  vnode = (VisitNode)enum.nextElement();
	  openList.addNode(vnode.nodeCost, vnode);
	}

/* ??? reset the visited list (want new visits to replace the old ones) */
	// visitedList = new Hashtable();

	setStatus("searching...");
	searchThread = new Thread(this);   // launch a thread to do the search
	searchThread.start();
      }
    }
    else if (source == bAddAllow) {
      /*
      String newCon = textNewCon.getText();
      if (newCon.equals("") == false) {
	allowedUrls.addElement(newCon);
	allowList.add(newCon);
	textNewCon.setText("");           // clear out entry box
      }
      else
	setStatus("Enter a new URL prefix before pressing this button");
	*/
      ConstraintAdder conAdder = new ConstraintAdder(allowList, allowedUrls);
      conAdder.show();
    }
    else if (source == bDelAllow) {
      String selCon = allowList.getSelectedItem();
      int selIndx = allowList.getSelectedIndex();
      if (selIndx >= 0) {
	allowList.delItem(selIndx);
	allowedUrls.removeElement(selCon);
      }
      else
	setStatus("Select from the Constraint List before " +
		  "pressing this button");
    }
    else if (source == bAddAvoid) {
      ConstraintAdder conAdder = new ConstraintAdder(avoidList, prohibUrls);
      conAdder.show();
    }
    else if (source == bDelAvoid) {
      String selCon = avoidList.getSelectedItem();
      int selIndx = avoidList.getSelectedIndex();
      if (selIndx >= 0) {
	avoidList.delItem(selIndx);
	prohibUrls.removeElement(selCon);
      }
      else
	setStatus("Select from the Constraint List before " +
		  "pressing this button");
    }
    else if (source == buttonExit) {
      exit();
    }
  }


  /** Adds the specified string to the main display window. */
  public void appendDetails(String details) {

    detailText.append(details);
  }


  public void exit(){

    Enumeration onts;

    /*
    System.out.println("Open List:");
    System.out.println(openList.toString());
    System.out.println();
    System.out.println("Closed List:");
    System.out.println(closedList.toString());
    */

    writeInitFile();
    writeVisitedFile();
    if (shoeKb != null) {
      shoeKb.closeKB();
      shoeKb.closeConnection();
    }
    logOut.close();
    System.exit(0);
  }


  /** Returns a KB interface suitable for editing an existing KB. This
      function can be overriden so that different backends can be
      used. */
  //change to abstract by Zpan on 5-30-02
  public abstract KBInterface getInterfaceForEditKb();
  /*{

    System.out.println("Using Parka for existing KB");
    return new ParkaInterface();    // set to update KB dynamically
  }*/


  /** Returns a KB interface suitable for creating a new KB. This
      function can be overriden so that different backends can be
      used. */
  //change to abstract by Zpan on 5-30-02
  public abstract KBInterface getInterfaceForNewKb();
  /*{

    System.out.println("Using Parka files for new KB");
    return new ParkaFileInterface();    // set to create new KB file
  }*/


  /** Returns a ShoeKb object which allows standard SHOE operations to be
    excuted against an underlying knowledge base. This function can be
    overriden so that a subclass of ShoeKb with an alternate implementation
    can be returned. */
  public ShoeKb getShoeKb(KBInterface kb, OntManager ontman,
			  PrintStream logout)
  {
    /* Note it is important to pass the protected implicitClaims
       variable so that ShoeKb knows whether or not to track
       and add them. */
    return new ShoeKb(kb, ontman, logout, implicitClaims);
  }


  /** Initializes a search by resetting search variables and adding the
    starting URL to the open list. */
  public void initializeSearch() {

    URL url;
    VisitNode vnode;

    updateUserParameters();
    openList = new SearchList();
    closedList = new Hashtable();

    try {
      url = new URL(startUrl);
    } catch (MalformedURLException ex) {
      setStatus(startUrl + " is not a valid URL");
      return;
    }
    // put URL in cannonical format
    url = normalizeURL(url);
    vnode = new VisitNode(url, 0, 1);
    openList.addNode(1, vnode);

    searchCount = 0;
    shoeCount = 0;
    errCount = 0;
    updateTextCounts();
  }


  /** Gets the contents of a URL for robot. Will not fetch URLs that are
      robot excluded (i.e., match an entry in the robots.txt file on the
      server) and will return an error if the server returns anything other
      than an OK response, even if HTML content is included (most servers
      return HTML error messages).
      */
  public InputStream robotFetch(String userAgent, String fromUser,
				URL url, Date lastVisited)
       throws IOException, RobotExcludedException, NotModifiedException {

    HttpURLConnection connect;
    int response;
    InputStream contentStream = null;

    if (robotEx.isExcluded(url))
      throw new RobotExcludedException();

    // System.out.println("HTTP Connect Date: " + httpDate.format(new Date()));
    connect = (HttpURLConnection)url.openConnection();
    connect.setRequestMethod("GET");
    connect.setRequestProperty("From", fromUser);
    connect.setRequestProperty("User-Agent", userAgent);

    // if using a proxy server with a login, set the Proxy-Authorization
    if (useProxy == true && proxyLogin.equals("") == false)
      connect.setRequestProperty("Proxy-Authorization", proxyAuth);

    if (lastVisited != null) {
      /* System.out.println("Setting If-Modified-Since to " +
			 httpDate.format(lastVisited));
	 */
      connect.setRequestProperty("If-Modified-Since",
				 httpDate.format(lastVisited));
    }
    response = connect.getResponseCode();
    if (response == HttpURLConnection.HTTP_NOT_MODIFIED)
      throw new NotModifiedException();
    else if (response != HttpURLConnection.HTTP_OK)
      throw new IOException();
    // System.out.println("Response code = " + response);
    // System.out.println("Length = " + connect.getContentLength());
    /*
    String hfield = connect.getHeaderField("Date");
    if (hfield != null)
      System.out.println("Header Field: Date = " + hfield);
      */
    contentStream = (InputStream)connect.getContent();
    return contentStream;
  }


  /** The main control loop of the web crawler. */
  public void run() {

    VisitNode vnode;
    Date startTime, endTime;
    SHOE_Doc doc;
    InputStream inStream;
    String detailLine;
    URL url;
    int cost = 0;
    boolean hasSHOEerrs;              // true if the document has SHOE errors
    boolean loadedDoc;                // true if a document was loaded
    String statusMsg;

    startTime = new Date();
    while ((goalSatisfied(cost) == false)
	   && (Thread.currentThread() == searchThread)) {

      vnode = chooseAction();               // get the next page to process
      cost = vnode.nodeCost;

      /* Retrieve page */
      /* Note we don't print the log information until after the fetch
	 because if the fetch is to a new server, we must read a new robots.txt
	 and print a message that indicates that we did */
      doc = new SHOE_Doc();
      url = vnode.url;
      detailLine = formatSearchDetail(url.toString());
      hasSHOEerrs = false;
      loadedDoc = false;
      statusMsg = "";

      try {
	inStream = robotFetch(userAgent, fromUser, url, vnode.lastVisited);
	doc.parse(inStream);
	inStream.close();
	loadedDoc = true;
      } catch (IOException ex) {
	statusMsg = "ERROR READING URL";
	errCount++;
	loadedDoc = false;
      } catch (RobotExcludedException ex) {
	statusMsg = "ROBOT EXCLUDED";
	loadedDoc = false;
      } catch (NotModifiedException ex) {
	statusMsg = "NOT MODIFIED";
	loadedDoc = false;
	/* update last visit date for the node. Note, to get a
	   NotModifiedException we must have previously visited the node */
	vnode = (VisitNode)visitedList.get(url);
	if (vnode != null)                // if null, then something very wrong
	  vnode.setLastVisited(new Date());
      }

      if (loadedDoc == true) {
	shoeKb.storeSHOEDoc(doc, vnode.url.toString());
	updateVisitedList(doc, vnode);
	expandNode(doc, vnode);

	/* this line must occur after storeSHOEDoc() b/c some SHOE errors
	   won't be detected until the document is validated in that
	   method. */
	if (doc.hasParseErrors()) {
	  statusMsg = "SHOE: HAS ERRORS";
	  hasSHOEerrs = true;
	}
	else if (doc.getVersion() == 1.0) {
	  statusMsg = "SHOE";
	  shoeCount++;                         // add to count of SHOE pages
	}
	else {
	  statusMsg = "";
	}
      }

      /* Print out the status and log information. By saving it all til
       the end, we can ensure proper ordering of the messages. */
      detailLine = detailLine + "\t" + statusMsg;
      appendDetails(detailLine + "\n");
      updateTextCounts();

      logOut.println(detailLine);
      // print messages from storing document (e.g. ontology load messages)
      logOut.print(shoeKb.getDocMessages());
      // print parse errors if any
      if (hasSHOEerrs == true) {
	logOut.println("Parse Errors:");
	logOut.println(doc.getParseErrors());
      }

      System.out.println(url.toString() + "\t" + vnode.parentCost +
			 "\t" + vnode.nodeCost);

      /* be nice to servers and wait before requesting another page */
      try {
	Thread.sleep(requestInterval);
      } catch (InterruptedException ex) {
	// continue if interrupted
      }
    }

    buttonSearch.setEnabled(true);
    buttonRevisit.setEnabled(true);
    buttonStop.setEnabled(false);
    buttonResume.setEnabled(true);

    updateTextCounts();

    /* if search is done, close the KB */
    if (goalSatisfied(cost) == true) {
      shoeKb.closeKB();
      shoeKb.closeConnection();
    }

    if (searchCount >= maxPages)
      setStatus("Reached search limit of " + maxPages);
    else
      setStatus("Done");
    endTime = new Date();

    /* print completion time information */
    long duration = endTime.getTime() - startTime.getTime();
    detailLine = "Search took " + duration + " ms\n";
    appendDetails(detailLine);
    logOut.println(detailLine);

    searchThread = null;
  }


  /** Sets up the frame which displays information about the web-crawler's
    progress and allows the user to interact with it. */
  public void setup() {

    Label tempLabel;
    Enumeration enum;

    setBackground(Color.lightGray);

    // set up the main UI panel
    panelMain = new Panel();
    panelMain.setLayout(new GridBagLayout());

    // text entry components
    tempLabel = new Label("Starting URL: ", Label.RIGHT);
    textURL = new TextField(startUrl, 60);
    textURL.setBackground(Color.white);
    Utilities.placeComponent(panelMain, tempLabel, 0, 0, 2, 1);
    Utilities.placeComponent(panelMain, textURL, 2, 0, 6, 1);
    tempLabel = new Label("KB Name: ", Label.RIGHT);
    textKB = new TextField(kbName, 60);
    textKB.setBackground(Color.white);
    Utilities.placeComponent(panelMain, tempLabel, 0, 1, 2, 1);
    Utilities.placeComponent(panelMain, textKB, 2, 1, 6, 1);

    tempLabel = new Label("Visit URL Prefixes: ", Label.RIGHT);
    allowList = new List(5);
    allowList.setBackground(Color.white);
    Utilities.placeComponent(panelMain, tempLabel, 0, 2, 2, 1);
    Utilities.placeComponent(panelMain, allowList, 2, 2, 5, 3);
    enum = allowedUrls.elements();
    while (enum.hasMoreElements())
      allowList.add((String)enum.nextElement());

    tempLabel = new Label("Avoid URL Prefixes: ", Label.RIGHT);
    avoidList = new List(5);
    avoidList.setBackground(Color.white);
    Utilities.placeComponent(panelMain, tempLabel, 0, 5, 2, 1);
    Utilities.placeComponent(panelMain, avoidList, 2, 5, 5, 3);
    enum = prohibUrls.elements();
    while (enum.hasMoreElements())
      avoidList.add((String)enum.nextElement());

    // tempLabel = new Label("New Constraint: ", Label.RIGHT);
    // textNewCon = new TextField("", 60);
    bAddAllow = new Button("Add");
    bDelAllow = new Button("Delete");
    bAddAvoid = new Button("Add");
    bDelAvoid = new Button("Delete");
    bAddAllow.addActionListener(this);
    bDelAllow.addActionListener(this);
    bAddAvoid.addActionListener(this);
    bDelAvoid.addActionListener(this);
    // Utilities.placeComponent(panelMain, tempLabel, 0, 5, 2, 1);
    // Utilities.placeComponent(panelMain, textNewCon, 2, 5, 6, 1);
    Utilities.placeComponent(panelMain, bAddAllow, 7, 2, 2, 1);
    Utilities.placeComponent(panelMain, bDelAllow, 7, 3, 2, 1);
    Utilities.placeComponent(panelMain, bAddAvoid, 7, 5, 2, 1);
    Utilities.placeComponent(panelMain, bDelAvoid, 7, 6, 2, 1);

    // list of result URLs
    Label labelResults = new Label("Search results");
    Utilities.placeComponent(panelMain, labelResults, 0,7,1,1);

    totCntText = new TextField("0", 5);
    shoeCntText = new TextField("0", 5);
    errCntText = new TextField("0", 5);
    totCntText.setBackground(Color.lightGray);
    shoeCntText.setBackground(Color.lightGray);
    errCntText.setBackground(Color.lightGray);
    totCntText.setEditable(false);
    shoeCntText.setEditable(false);
    errCntText.setEditable(false);
    Utilities.placeComponent(panelMain, new Label("URL Count:"), 1,8,1,1);
    Utilities.placeComponent(panelMain, totCntText, 2,8,1,1);
    Utilities.placeComponent(panelMain, new Label("SHOE Count:"), 3,8,1,1);
    Utilities.placeComponent(panelMain, shoeCntText, 4,8,1,1);
    Utilities.placeComponent(panelMain, new Label("Error Count:"), 5,8,1,1);
    Utilities.placeComponent(panelMain, errCntText, 6,8,1,1);

    detailText = new TextArea(15, 90);
    detailText.setBackground(Color.lightGray);
    detailText.setEditable(false);
    detailText.setFont(new Font("Monospaced", Font.PLAIN, 8));
    Utilities.placeComponent(panelMain, detailText, 0,9,8,6, 1.0,1.0);

    labelStatus = new Label("");
    Utilities.placeComponent(panelMain, labelStatus, 0,16,8,1, 1.0,0.0);

    // control buttons
    Panel panelButtons = new Panel();
    buttonSearch = new Button(SEARCH);
    buttonRevisit = new Button(REVISIT);
    buttonStop = new Button(STOP);
    buttonResume = new Button(RESUME);
    buttonExit = new Button(EXIT);
    buttonSearch.addActionListener(this);
    buttonRevisit.addActionListener(this);
    buttonStop.addActionListener(this);
    buttonResume.addActionListener(this);
    buttonExit.addActionListener(this);
    buttonStop.setEnabled(false);
    buttonResume.setEnabled(false);
    panelButtons.add(buttonSearch);
    panelButtons.add(buttonRevisit);
    panelButtons.add(buttonStop);
    panelButtons.add(buttonResume);
    panelButtons.add(buttonExit);

    Utilities.placeComponent(panelMain, panelButtons, 0,17,8,1, 1.0,0.0);
    add(panelMain);
  }


  /** Sets the status field in the display to the specified string. */
  public void setStatus(String status) {

    labelStatus.setText(status);
  }


  /** Sets user configurable parameters based upon the users
    entries into text fields. This method should be called
    anytime these parameters are to be used. */
  public void updateUserParameters() {

    // this may be called before the text widgets are initialized
    if (textKB != null) {
      kbName = textKB.getText();
      startUrl = textURL.getText();
    }
  }

  /** Updates the display with the current counts for total pages
    visited, SHOE pages found, and errors found. */
  public void updateTextCounts() {

    totCntText.setText(String.valueOf(searchCount));
    shoeCntText.setText(String.valueOf(shoeCount));
    errCntText.setText(String.valueOf(errCount));
  }



  /* ********************* Protected methods *********************** */


  /** Selects the next action to perform. Actions are kept in a queue.
     After an action is chosen, it is put on a closed list so that
     it will never be repeated. */
  protected VisitNode chooseAction() {

    VisitNode vnode;
    URL url;

    vnode = (VisitNode)openList.getNextNode();
    url = vnode.url;
    closedList.put(url, "");          // put node on the closed list
    searchCount++;
    return vnode;
  }


  /** Add all hypertext links (<A> tags) to the list of pages to visit. */
  protected void expandHypertext(Enumeration htkns, VisitNode parent) {

    HTML_Token token;
    SGML_Tag tag;
    String link;

    while (htkns.hasMoreElements()) {
      token = (HTML_Token)htkns.nextElement();
      if (token instanceof SGML_Tag) {
	tag = (SGML_Tag)token;
	if (tag.getName().equals("A")) {                // hypertext link tag
	  link = tag.getAttribute("HREF");
	  if (link != null)
	    queueNode(link, parent);
	}
      }
    }
  }


  /** Finds all web pages reachable from the given document, either by
    hypertext or by SHOE arguments. For each URL, if it is not on
    the closed list, then it is inserted into the appropriate place
    on the open list. */
  protected void expandNode(SHOE_Doc doc, VisitNode parent) {

    Enumeration enum;
    SHOE_Obj docContent;
    SHOE_Obj obj;
    Instance inst;
    Enumeration elts;
    String forText;

    Relation rel;
    Def_Relation drel;
    Def_Arg dargs[];
    int typeid;
    RelArg arglist[];
    RelArg arg;
    int argcnt;
    int i;
    String argval;

    expandHypertext(doc.getHTML(), parent);

    /* examine claims about other instances */
    if (doc.getVersion() == 1.0) {
      enum = doc.getPSHOE();
      while (enum.hasMoreElements()){
	docContent = (SHOE_Obj)enum.nextElement();
	if (docContent instanceof Instance) {
	  inst = (Instance)docContent;
	  /* note that the instance should have been validated when we
	     called ShoeKb.storeSHOEDoc() */
	  elts = inst.getDeclarations();
	  while (elts.hasMoreElements()) {
	    obj = (SHOE_Obj)elts.nextElement();
	    /* look at instances that are arguments in relations */
	    if (obj instanceof Relation) {
	      rel = (Relation)obj;
	      drel = (Def_Relation)ontman.getElement(rel.getElmtId());
	      if (drel != null) {
		dargs = drel.getDefArgs();
		arglist = rel.getArgs();
		argcnt = arglist.length;
		for (i=0; i < argcnt; i++) {
		  typeid = dargs[i].getTypeClass();
		  /*
		    System.out.println ("Arg " + i + ", type = " +
		    dargs[i].getType() + ": typeid = " + typeid);
		    */
		  if (typeid == Def_Arg.CATEGORY) {
		    /* if argument is an instance, and does not describe the
		       current instance, then consider it */
		    arg = arglist[i];
		    argval = arg.getValue();
		    if (argval != null) {       // args may have null values
		      if (argval.equals(inst.getKey()) == false)
			/* only consider as a link if it starts with http://
			   (instance keys can not be relative URLs */
			if (argval.startsWith("http://") == true)
			  queueNode(argval, parent);
		    }
		  }
		}
	      }
	      else
		;     // ignore invalid relations
	    }
	    /* look at instances that are claimed to be of another category */
	    if (obj instanceof Category) {
	      forText = ((Category)obj).getForText();
	      /* if category for some other instance, add that instance */
	      if (forText.equals(inst.getKey()) == false)
		queueNode(forText, parent);
	    }
	  }
	}
      }
    }
  }


  /** Given a URL, returns a string that contains the URL and the
    current time, and pads the space between them with tabs in
    an attempt to achieve distinct columns for each when used to
    displaying the web-crawler's progress. */
  protected String formatSearchDetail(String url) {

    String line;
    int len;
    int spaces;
    int tabs;
    int i;

    len = url.length();
    spaces = 63 - len;  // use a multiple of 8, minus 1 to ensure good columns
    if (spaces > 0)
      tabs = (spaces / TABSIZE) + 1;
    else
      tabs = 1;
    line = url;
    for (i=0; i < tabs; i++)
      line = line + "\t";
    line = line + dformat.format(new Date());
    // System.out.println("Len=" + len + ", Spaces=" + spaces + ", tabs= " + tabs);
    return line;
  }


  /** Given a url and it's parent node, estimates the cost of visting the
     node in terms of its likelyhood to contain SHOE. URLs that don't
     end in 'html' and URLs that are outside of the user specified
     constraints are given infinite costs. URLs that are in different
     directories than their parents are more expensive than those that
     aren't. (This assumes that SHOE is localized). */
  protected int getCostEstimate(URL url, VisitNode parent) {

    int cost;
    String host;
    String file;

    cost = parent.nodeCost;

    host = url.getHost();
    file = url.getFile();

    // System.out.println("Host: " + host + ", File: " + file);

    /* if ends with extension other than html/htm, then never visit */
    if (file.indexOf('.') > 0)
      if (file.endsWith(".htm") == false &&
	  file.endsWith(".html") == false)
	cost = maxCost + 1;

    if (meetsConstraints(url.toString()) == false)
      cost = maxCost + 1;

    // Increase cost if new node is in a different directory
    if (inSameDirectory(parent.url, url) == false)
      cost = cost + C_DIFF_DIR;

    return cost;
  }


  /** Returns the cost of visiting the page; pages with lower costs will
    be visited first. */
  protected int getPageCost(SHOE_Doc doc, int parentCost) {

    int cost = 0;

    if (doc.getPSHOE().hasMoreElements() == true)
      cost = parentCost + C_SHOE_PAGE;
    else
      cost = parentCost + C_HTML_PAGE;
    return cost;
  }


  /** Returns true if the agent has reached a termination criteria. */
  protected boolean goalSatisfied(int cost) {

    if (openList.isEmpty() ||
	cost >= maxCost ||
	searchCount >= maxPages)
      return true;
    else
      return false;
  }


  /** Returns true if the second URL is in the same directory or a
    subdirectory of the first URL. */
  protected boolean inSameDirectory(URL parUrl, URL childUrl) {

    String parFile;                 // parent file name (from URL)
    String childFile;
    int dirEnd;                     // index where directory ends in parFile
    String parDir;                  // parent directory name

    parFile = parUrl.getFile();
    childFile = childUrl.getFile();
    dirEnd = parFile.lastIndexOf('/');
    parDir = parFile.substring(0, dirEnd);
    return childFile.startsWith(parDir);
  }


  /** Configures the system to use a proxy server when fetching
    URLs. */
  protected void initProxyServer(String proxyLogin, String proxyHost,
				 String proxyPort) {

    proxyAuth = "Basic " + new sun.misc.BASE64Encoder().encode(proxyLogin.getBytes());
    System.getProperties().put("proxySet", "true");
    System.getProperties().put("proxyHost", proxyHost);
    System.getProperties().put("proxyPort", proxyPort);
  }


  /** Returns true if the URL is within one of the locations specified
    by the user's URL constraints. Specifically, the URL must begin
    with one of the prefixes specified in the allowed URLs and must
    not begin with any of the prefixes specified as prohibited URLs. */
  protected boolean meetsConstraints(String urlStr) {

    Enumeration enum;
    String cnstrnt;
    boolean satisfy = false;

    enum = allowedUrls.elements();
    while (enum.hasMoreElements() && satisfy == false) {
      cnstrnt = (String)enum.nextElement();
      if (urlStr.startsWith(cnstrnt))
	satisfy = true;
    }
    if (satisfy == true) {
      enum = prohibUrls.elements();
      while (enum.hasMoreElements() && satisfy == true) {
	cnstrnt = (String)enum.nextElement();
	if (urlStr.startsWith(cnstrnt))
	  satisfy = false;
      }
    }
    return satisfy;
  }


  /** Given a URL, converts it into a cannonical form. For our purposes, this
     involves adding a '/' after URLs that don't have a file extension,
     removing index.html from the URL, converting '~' to 'user/', and
     removing any references to named anchors from the end of the URL.
     Note: these methods may not always be correct, we need a way to
     use the old URL if the normalized one doesn't work.
     */
  protected static URL normalizeURL(URL url) {

    String protocol;
    String host;
    int port;
    String file;
    String ref;
    int pathEnd;
    String filename;
    int tildeLoc;

    protocol = url.getProtocol();
    host = url.getHost();
    port = url.getPort();
    file = url.getFile();
    ref = url.getRef();

    pathEnd = file.lastIndexOf('/');
    if (pathEnd >= 0)
      filename = file.substring(pathEnd + 1);
    else
      filename = file;

    /* Does URL have a file extension? */
    if (filename.indexOf('.') >= 0) {
      if (filename.equalsIgnoreCase("index.htm") ||
	  filename.equalsIgnoreCase("index.html"))
	file = file.substring(0, pathEnd + 1);       // don't include filename
    }
    else {
      /* assume that URLs without a file extension are directories where the
	 trailing slash was forgotten */
      if (file.endsWith("/") == false)
	file = file + "/";
    }

    /* expand the tilde to 'users/' */
    /* DISABLED - this doesn't always work */
    /*
    tildeLoc = file.indexOf('~');
    if (tildeLoc >= 0) {
      file = file.substring(0, tildeLoc) + "users/" +
	file.substring(tildeLoc + 1);
    }
    */

    /* when we create the new URL, we intentionaly leave out the reference */
    try {
      url = new URL(protocol, host, port, file);
    } catch (MalformedURLException ex) {
      // setStatus("Couldn't create new URL");
    }
    return url;
  }


  /** Given a string representing a URL, estimates the value of visiting
    it and inserts it into the the appropriate place in the queue.
    */
  protected void queueNode(String link, VisitNode parent) {

    URL context;
    int parentCost;
    int fitness = 0;
    URL url;
    VisitNode vnode;

    context = parent.url;
    parentCost = parent.nodeCost;

    try {
      /* construct new URL, this may be a relative URL */
      /* if it is, the context must end with a '/' or a file name for
	 the URL to be created properly */

      url = new URL(context, link);
      url = normalizeURL(url);

      /* if link not on closed list, add it to open list */
      if (closedList.containsKey(url) == false) {
	fitness = getCostEstimate(url, parent);
	/* if cost too high, then no need to even put in queue */
	if (fitness > maxCost)
	  closedList.put(url, "");          // put node on the closed list
	else {
	  vnode = new VisitNode(url, parentCost, fitness);
	  openList.addNode(fitness, vnode);
	}
      }
    } catch (MalformedURLException ex) {
      /* nothing is done here because this routine is used to filter out
	 the things that are not URLs */
      ; // System.out.println(link + " is not a valid URL");
     }
  }


  /** Reads the initialization file which consists of the starting URL and
     a list of valid server/path prefixes for searching. */
  /* Note we do not use the Java 1.1. class Properties because we need
     some properties that have multiple values (i.e., allowedUrls and
     prohibUrls). */
  protected boolean readInitFile() {

    BufferedReader initInt = null;
    boolean fileExists = true;
    String inLine;
    int dataIndx;

    // load last information
    try{
      initInt = new BufferedReader(new FileReader(initFile));
    } catch (FileNotFoundException ex) {
      // if file not found, don't try to load visited list
      System.out.println("Couldn't find property file " + initFile);
      fileExists = false;
    }
    if (fileExists) {
      try {
	while((inLine=initInt.readLine()) != null) {   // read a line
	  if (inLine.startsWith("FROM_USER="))
	    fromUser = extractPropertyStringValue(inLine);
	  else if (inLine.startsWith("KB_HOST="))
	    kbHost = extractPropertyStringValue(inLine);
	  else if (inLine.startsWith("KB_PORT="))
	    kbPort = extractPropertyIntValue(inLine);
	  else if (inLine.startsWith("KBNAME="))
	    kbName = extractPropertyStringValue(inLine);
	  else if (inLine.startsWith("START_URL="))
	    startUrl = extractPropertyStringValue(inLine);
	  else if (inLine.startsWith("ALLOW_URL="))
	    allowedUrls.addElement(extractPropertyStringValue(inLine));
	  else if (inLine.startsWith("PROHIB_URL="))
	    prohibUrls.addElement(extractPropertyStringValue(inLine));
	  else if (inLine.startsWith("MAX_PAGES="))
	    maxPages = extractPropertyIntValue(inLine);
	  else if (inLine.startsWith("MAX_COST="))
	    maxCost = extractPropertyIntValue(inLine);
	  else if (inLine.startsWith("REQUEST_INTERVAL="))
	    requestInterval = extractPropertyIntValue(inLine);
	  else if (inLine.startsWith("IMPLICIT_CLAIMS="))
	    implicitClaims = extractPropertyBooleanValue(inLine);
	  else if (inLine.startsWith("USE_PROXY="))
	    useProxy =  extractPropertyBooleanValue(inLine);
	  else if (inLine.startsWith("PROXY_LOGIN="))
	    proxyLogin = extractPropertyStringValue(inLine);
	  else if (inLine.startsWith("PROXY_HOST="))
	    proxyHost = extractPropertyStringValue(inLine);
	  else if (inLine.startsWith("PROXY_PORT="))
	    proxyPort = extractPropertyStringValue(inLine);
	  else {       // handle old format for allowed URLs
	    allowedUrls.addElement(inLine);
	  }
	}
	initInt.close();
      }catch (IOException e) {
	e.printStackTrace();
      }
    }
    return fileExists;
  }


  /** Reads the file containing information about URLs previously
    visited and updates the visited list appropriately. */
  protected boolean readVisitedFile() {

    BufferedReader visitIn = null;
    boolean fileExists = true;
    String inLine;
    StringTokenizer tokens;
    String urlStr;
    String dateStr;
    String scoreStr;
    URL url = null;
    Date lastVisited = null;
    int score = 0;
    VisitNode vnode;

    // load last information
    try{
      visitIn = new BufferedReader(new FileReader(visitFile));
    } catch (FileNotFoundException ex) {
      // if file not found, don't try to load visited list
      System.out.println("Couldn't find visited file " + visitFile);
      fileExists = false;
    }
    if (fileExists) {
      try {
	while((inLine=visitIn.readLine()) != null) {
	  tokens = new StringTokenizer(inLine, "\t");
	  urlStr = tokens.nextToken();
	  dateStr = tokens.nextToken();
	  scoreStr = tokens.nextToken();
	  try {
	    url = new URL(urlStr);
	    lastVisited = httpDate.parse(dateStr);
	    score = Integer.parseInt(scoreStr);
	  } catch (MalformedURLException ex) {
	    System.out.println(ex.getMessage());
	  } catch (ParseException ex) {
	    System.out.println(ex.getMessage());
	  } catch (NumberFormatException ex) {
	    System.out.println(ex.getMessage());
	  }
	  /* NOTE: we may need to change the file to store both costs */
	  vnode = new VisitNode(url, score, score, lastVisited);
	  visitedList.put(url, vnode);
	}
	visitIn.close();
      }catch (IOException e) {
	e.printStackTrace();
      }
    }
    return fileExists;
  }


  /** Determine the path cost including the current page and update the
    visited list. If the node has been visited before, modify the date
    and score. Otherwise, insert a new node. */
  protected void updateVisitedList(SHOE_Doc doc, VisitNode vnode) {

    int pageCost;
    VisitNode listNode;
    int parentCost;
    URL docURL;

    parentCost = vnode.parentCost;
    docURL = vnode.url;

    /* TEST: Try using the estimated cost as the base cost. This way,
       examining a new directory is more expensive. */
    // pageCost = getPageCost(doc, parentCost);
    pageCost = getPageCost(doc, vnode.nodeCost);

    /* update cost and date for current node. The cost will be used by
       later functions */
    vnode.setLastVisited(new Date());
    vnode.setNodeCost(pageCost);

    /* update visited list with information about this page */
    if (visitedList.containsKey(docURL)) {            // update list's node
      listNode = (VisitNode)visitedList.get(docURL);
      listNode.setLastVisited(new Date());
      listNode.setNodeCost(pageCost);
    }
    else {                            // add the current node to visited list
      visitedList.put(docURL, vnode);
    }
  }


  protected void writeInitFile() {

    PrintWriter initOut;
    String strURL;
    Enumeration enum;

    updateUserParameters();

    try {
      /* overwrite exisiting file b/c dates and scores may change */
      initOut = new PrintWriter(new FileOutputStream(initFile));

      initOut.println("FROM_USER=" + fromUser);
      initOut.println("KB_HOST=" + kbHost);
      initOut.println("KB_PORT=" + kbPort);
      initOut.println("KBNAME=" + kbName);
      initOut.println("START_URL=" + startUrl);
      enum = allowedUrls.elements();
      while (enum.hasMoreElements()) {
	initOut.println("ALLOW_URL=" + enum.nextElement());
      }
      enum = prohibUrls.elements();
      while (enum.hasMoreElements()) {
	initOut.println("PROHIB_URL=" + enum.nextElement());
      }
      initOut.println("MAX_PAGES=" + maxPages);
      initOut.println("MAX_COST=" + maxCost);
      initOut.println("REQUEST_INTERVAL=" + requestInterval);
      initOut.print("IMPLICIT_CLAIMS=");
      if (implicitClaims == true)
	initOut.println("TRUE");
      else
	initOut.println("FALSE");
      if (useProxy == true) {
	initOut.println("USE_PROXY=TRUE");
	initOut.println("PROXY_LOGIN=" + proxyLogin);
	initOut.println("PROXY_HOST=" + proxyHost);
	initOut.println("PROXY_PORT=" + proxyPort);
      }

      initOut.close();
    }catch (IOException e) {
      e.printStackTrace();
    }
  }

  protected void writeVisitedFile() {

    PrintWriter visitOut;
    String strURL;
    VisitNode vnode;
    Enumeration enum;

    try {
      /* overwrite exisiting file b/c dates and scores may change */
      visitOut = new PrintWriter(new FileOutputStream(visitFile));

      enum = visitedList.elements();
      while (enum.hasMoreElements()) {
	vnode = (VisitNode)enum.nextElement();
	visitOut.print(vnode.url + "\t");
	visitOut.print(httpDate.format(vnode.lastVisited) + "\t");
	visitOut.print(vnode.nodeCost + "\n");
      }
      visitOut.close();
    }catch (IOException e) {
      e.printStackTrace();
    }
  }



  /* ********************* Private methods *********************** */


  /** Given a string that represents a line from a property file,
    returns the value of the property. */
  private String extractPropertyStringValue(String line) {

    int    dataIndx;
    String value;

    dataIndx = line.indexOf("=") + 1;
    value = line.substring(dataIndx);
    return value;
  }


  private int extractPropertyIntValue(String line) {

    int    dataIndx;
    String strValue;
    int    intValue;

    dataIndx = line.indexOf("=") + 1;
    strValue = line.substring(dataIndx);
    try {
      intValue = Integer.parseInt(strValue);
    } catch (NumberFormatException ex) {
      System.out.println("In property line " + line + ", " + strValue +
			 " must be an integer.");
      intValue = 0;
    }
    return intValue;
  }


  private boolean extractPropertyBooleanValue(String line) {

    int dataIndx;
    String strValue;
    boolean booleanValue;

    strValue = extractPropertyStringValue(line);
    if (strValue.equalsIgnoreCase("TRUE"))
      booleanValue = true;
    else
      booleanValue = false;
    return booleanValue;
  }

  /** Loads the ontology proxy information from a local file. Each
    line specifies an ontology id, version and a proxy URL for
    retrieving it. These proxies are then set in the OntManager.
    */
  private void loadOntProxyFile() {

    // This occurs before the screen is setup, so all error messages
    // must be sent to System.err

    FileInputStream inStream;
    String errMsg;

    File proxyFile = new File(PROXY_FILE);
    try {
      inStream = new FileInputStream(proxyFile);
      errMsg = ontman.loadOntProxyInputStream(inStream);
      if (errMsg != null) {
	System.err.println("Found errors while reading ontology proxy file " +
		   PROXY_FILE + ":\n" + errMsg);
      }
      System.err.println("Loaded ont proxy file for application");
    } catch (FileNotFoundException e) {
      System.err.println("There is no ontology proxy file: " + PROXY_FILE);
    }
  }
}
