/* Copyright (C) 1999 Jeff Heflin - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
*/


/*change history
//05-29-02 change global variables to protected  -zpan

*/

package Expose;

import java.io.*;
import java.util.*;
import java.net.*;
import java.text.*;                // for formatting and parsing dates
import SHOE.*;
import ShoeKb.*;
import HTML.*;
import UserInt.Utilities;

/** An application that given a URL, fetches the corresponding web page
  and stores the information contained within a Parka knowledge base.

  USAGE:
  java Expose.SHOELoader url
  java Expose.SHOELoader url kbname host port
  */
public abstract class SHOELoader
{
  /* Constants */
  public static String DEFAULT_KB = null;
  public static String DEFAULT_HOST = null;
  public static int DEFAULT_PORT = 0;
  public static String logFile = "loader.log";
  public static boolean implicitClaims = true;

  /* Global vars */
  protected OntManager        ontman;             // used to cache ontologies
  protected ShoeKb            kb;                 // high-level interface to KB
  protected Hashtable         storedOnts;         // table of ontologies stored in KB

  protected DateFormat        dformat;
  protected SimpleDateFormat  httpDate;
  protected PrintStream       logOut = null;

  public static final int TABSIZE = 8;



  /** Constructor that causes the gathered data to be stored in the
    default knowledge base, located at the default host and port. */
  public SHOELoader(String startUrl)
  {
    if (DEFAULT_KB == null || DEFAULT_HOST == null) {
      System.err.println("Must set default KB, host and/or port in " +
			 "source code");
      System.exit(1);
    }
    beginLoader(DEFAULT_KB, DEFAULT_HOST, DEFAULT_PORT);
    processUrl(startUrl, DEFAULT_KB);
    exit(0);
  }

  /** Constructor that allows the user to specify a start URL, knowledge
    base, and the server name and port number through which the knowledge
    base can be accessed. */
  public SHOELoader(String startUrl, String kbName, String host, int port)
  {
    beginLoader(kbName, host, port);
    processUrl(startUrl, kbName);
    exit(0);
  }


  /** Constructor that allows a list of URLs to be specified in a
      file, and loads the SHOE from the locations into the default
      knowledge base, located at the defaul host and port. */
  public SHOELoader(File urlFile)
  {
    DataInputStream inStream;
    boolean eof = false;
    String line;
    String urlString;

    if (DEFAULT_KB == null || DEFAULT_HOST == null) {
      System.err.println("Must set default KB, host and/or port in " +
			 "source code");
      System.exit(1);
    }
    beginLoader(DEFAULT_KB, DEFAULT_HOST, DEFAULT_PORT);
    try {
      inStream = new DataInputStream(new FileInputStream(urlFile));
      while (!eof) {
	line = inStream.readLine();
	if (line == null || line.equals(""))
	  eof = true;
	else {
	  urlString = line;
	  processUrl(urlString, DEFAULT_KB);
	}
      }
    } catch (IOException ex) {
      System.out.println(ex.getMessage());
    }
    exit(0);
  }

  /** Constructor that allows a list of URLs to be specified in a
      file, and loads the SHOE from the locations into a specified
      knowlede base located at a specified server name and port number. */
  public SHOELoader(File urlFile, String kbName, String host, int port)
  {
    DataInputStream inStream;
    boolean eof = false;
    String line;
    String urlString;

    beginLoader(kbName, host, port);
    try {
      inStream = new DataInputStream(new FileInputStream(urlFile));
      while (!eof) {
	line = inStream.readLine();
	if (line == null || line.equals(""))
	  eof = true;
	else {
	  urlString = line;
	  processUrl(urlString, kbName);
	}
      }
    } catch (IOException ex) {
      System.out.println(ex.getMessage());
    }
    exit(0);
  }


  /** Initializes the loader object. */
  protected abstract void beginLoader(String kbName, String host, int port);

  protected void processUrl(String startUrl, String kbName) {

    URL url = null;

    logOut.print(formatSearchDetail(startUrl));

    try {
      url = new URL(startUrl);
    } catch (MalformedURLException ex) {
      logOut.println("\tINVALID URL");
      exit(1);
    }

    /*
    // put URL in cannonical format
    url = normalizeURL(url);
    */
    internPage(url, kbName);
  }


  /** Clean up before exiting by closing the knowledge base and the
    log files. */
  public void exit(int code) {

    Enumeration onts;

    if (kb != null) {
      kb.closeKB();
    }
    if (logOut != null)
      logOut.close();
    System.exit(code);
  }

  /** Given a URL, converts it into a cannonical form. For our purposes, this
     involves adding a '/' after URLs that don't have a file extension,
     removing index.html from the URL, converting '~' to 'user/', and
     removing any references to named anchors from the end of the URL.
     Note: these methods may not always be correct, we need a way to
     use the old URL if the normalized one doesn't work.
     */
  /* This procedure is not used by the single page loader */
  protected static URL normalizeURL(URL url) {

    String protocol;
    String host;
    int port;
    String file;
    String ref;
    int pathEnd;
    String filename;
    int tildeLoc;

    protocol = url.getProtocol();
    host = url.getHost();
    port = url.getPort();
    file = url.getFile();
    ref = url.getRef();

    pathEnd = file.lastIndexOf('/');
    if (pathEnd >= 0)
      filename = file.substring(pathEnd + 1);
    else
      filename = file;

    /* Does URL have a file extension? */
    if (filename.indexOf('.') >= 0) {
      if (filename.equalsIgnoreCase("index.htm") ||
	  filename.equalsIgnoreCase("index.html"))
	file = file.substring(0, pathEnd + 1);       // don't include filename
    }
    else {
      /* assume that URLs without a file extension are directories where the
	 trailing slash was forgotten */
      if (file.endsWith("/") == false)
	file = file + "/";
    }

    /* expand the tilde to 'users/' */
    /* temporarilly disabled because this isn't always true!!! */
    /*
    tildeLoc = file.indexOf('~');
    if (tildeLoc >= 0) {
      file = file.substring(0, tildeLoc) + "users/" +
	file.substring(tildeLoc + 1);
    }
    */

    /* when we create the new URL, we intentionaly leave out the reference */
    try {
      url = new URL(protocol, host, port, file);
    } catch (MalformedURLException ex) {
      // setStatus("Couldn't create new URL");
    }
    return url;
  }

  /** Retrieve a page and store any SHOE information in the KB. */
  public void internPage(URL url, String kbName) {

    SHOE_Doc doc;
    InputStream inStream;
    String detailLine = "";
    boolean hasSHOEerrs = false;

    /* Note we don't print the log information until after the fetch
       because if the fetch is to a new server, we must read a new robots.txt
       and print a message that indicates that we did */
    doc = new SHOE_Doc();
    try {
      inStream = url.openStream();
      doc.parse(inStream);
      inStream.close();
    } catch (IOException ex) {
      detailLine = detailLine + "\tERROR READING URL";
      logOut.println(detailLine);
      System.out.println("***************************************");
      System.out.println("Load status:" + detailLine);
      System.out.println("***************************************");
      return;
    }

    kb.storeSHOEDoc(doc, url.toString());

    /* this line must occur after storeSHOEDoc() b/c some SHOE errors
       won't be detected until the document is validated in that
       method. */
    if (doc.hasParseErrors()) {
      hasSHOEerrs = true;
      detailLine = detailLine + "\tSHOE: HAS ERRORS";
      logOut.print(detailLine);
    }
    else if (doc.getVersion() == 1.0) {
      detailLine = detailLine + "\tSHOE";
      logOut.print(detailLine);
    }
    else {
      detailLine = detailLine + "\tNOT SHOE";
      logOut.print(detailLine);
    }
    logOut.println();

    /* the following message is so that the CGI script will report the
       status of the load */
    System.out.println("***************************************");
    System.out.println("Load status:" + detailLine);

    if (hasSHOEerrs == true) {
      System.out.println();
      System.out.println("Parse Errors:");
      System.out.println(doc.getParseErrors());
    }
    System.out.println("***************************************");

  }

  /** Creates a string that provides status information and which has
    been formated for columnar display. */
  String formatSearchDetail(String url) {

    String line;
    int len;
    int spaces;
    int tabs;
    int i;

    len = url.length();
    spaces = 47 - len;  // use a multiple of 8, minus 1 to ensure good columns
    if (spaces > 0)
      tabs = (spaces / TABSIZE);
    else
      tabs = 0;
    line = dformat.format(new Date());
    line = line + "\t" + url;

    for (i=0; i < tabs; i++)
      line = line + "\t";

    // System.out.println("Len=" + len + ", Spaces=" + spaces + ", tabs= " + tabs);
    return line;
  }

}


