/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in ShoeKb.java
 * and the file License.html for details.
 */

/* Change History:
   - 09/28/00, JDH: Added the getVariables() method.
   */
package ShoeKb;

import java.util.*;

/** A logical formula. This formula may be asserted to a knowledge base
  or may be a query to a knowledge base. Currently, a sentence can consist
  of one or more atoms conjuncted together, where an atom is a predicate
  with one or more arguments. An argument may be a literal or a variable,
  where variables are indicated by preceding them with a question mark. */
public class Sentence {

  Vector atoms;

  /** Construct a new Sentence without initializing it. */
  public Sentence() {

    atoms = new Vector();
  }

  /** Construct a new Sentence which consists of a single predicate with
    any number of arguments. */
  public Sentence(String predicate, String[] args) {
    
    atoms = new Vector();
    addConjunct(predicate, args);
  }

  /** Constructs a new Sentence which consists of a single predicate with
    two arguments. */
  public Sentence(String predicate, String arg1, String arg2) {

    atoms = new Vector();
    addConjunct(predicate, arg1, arg2);
  }

  /** Adds an additional atom as a conjunct to the existing sentence
    where the conjunct consists of a predicate and a set of
    arguments. */
  public void addConjunct(String predicate, String[] args) {

    Atom atom;

    atom = new Atom(predicate, args);
    atoms.addElement(atom);
  }

  /** Adds an additional atom as a conjunct to the existing sentence
    where the atom consists of a predicate and two arguments. */
  public void addConjunct(String predicate, String arg1, String arg2) {

    String args[];

    args = new String[2];
    args[0] = arg1;
    args[1] = arg2;
    addConjunct(predicate, args);
  }

  /** Adds an additional atom as a conjunct to the existing sentence
    where the input is the atom to be added. */
  public void addConjunct(Atom atom) {

    atoms.addElement(atom);
  }

  /** Returns the vector of atoms that are conjuncted together
    in the sentence. */
  public Vector getAtoms() {

    return atoms;
  }


  /** Returns an array of the variables in the sentence. They are
    returned in the order in which they appear in the sentence. */
  public String[] getVariables() {

    Enumeration atomEnum;
    Atom        anAtom;
    String[]    args;
    Vector      varSet;
    String[]    varArray;

    varSet = new Vector();
    atomEnum = getAtoms().elements();
    while (atomEnum.hasMoreElements()) {
      anAtom = (Atom)atomEnum.nextElement();
      args = anAtom.getArguments();
      for (int i=0; i < args.length; i++) {
	if (args[i].startsWith("?")) {
	  if (varSet.contains(args[i]) == false)
	    varSet.addElement(args[i]);
	}
      }
    }
    varArray = new String[varSet.size()];
    varSet.copyInto(varArray);
    return varArray;
  }


  public String toString() {

    Enumeration conjuncts;
    Atom   atom;
    StringBuffer sbuf;

    sbuf = new StringBuffer();
    conjuncts = getAtoms().elements();
    while (conjuncts.hasMoreElements()) {
      atom = (Atom)conjuncts.nextElement();
      sbuf.append(atom.toString());
    }
    return sbuf.toString();
  }

}

