/* SHOE KB Libary
 * Copyright (C) 1999 Jeff Heflin - University of Maryland
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Jeff Heflin
 * Contact: Jeff Heflin (heflin@cs.umd.edu)
 *
 */

/* Change History:
   05/10/02 ZPAN - Add protected method setdocMessages()
   05/10/02 ZPAN - Add public methods getdefaultOntId() and getdefaultOntVer()
   01/29/01 JDH - Use new version of Instance.getUseOnts()
   01/04/01 JDH - Ontology catalogs now work when no URL is given for an ont
   01/02/01 JDH - Changed getUniqueName() to be more efficient
            JDH - Created toKbQuery() method, issueQuery() now calls it
   12/29/00 JDH - Changed a number of methods to protected
            JDH - Changed return type of getKnownOntologies()
	    JDH - added setBackwardCompatability() and setExtendedOnts()
	    JDH - changed type of argument to storeNewOntologies()
   11/03/00 JDH - Replaced Comparison.getType() with getOperator()
   11/01/00 JDH - Used accesor methods for fields in Def_Constant and
                  Def_Category
   10/26/00 JDH - Used accessor methods for fields in Category and Def_Relation
   05/24/00 JDH - Made issueQuery() pass on the numAnswersRequested to the KB
   05/08/00 JDH - Made sure the the "name" relation is associated with the
                  base ontology.
	    JDH - Automatically store the base ontology when a KB is created
	    JDH - Added a getUniqueName() that takes ont as an argument
   */

package ShoeKb;

import java.util.*;
import java.io.PrintStream;
import SHOE.*;

/** A high-level interface for communicating with a SHOE knowledge
  base. */
public class ShoeKb
{

  static final String BASE_ONT_ID = "base-ontology";
  static final String BASE_ONT_VER = "1.0";

  /* Global vars */
  protected OntManager   ontman;        // used to cache ontologies
  protected KBInterface  kb;            // stores interface to knowledge base
  protected Hashtable    storedOnts;    // table of ontologies stored in KB

  protected int maxSeqNum = 0;    // last seq. number assigned to an ontology

  PrintStream  logOut;                  // for output logging
  String       docMessages;             // messages about current document
  boolean      implicitClaims = false;  // if true, make implied claims

  String       defaultOntId  = BASE_ONT_ID;
  String       defaultOntVer = BASE_ONT_VER;
  String       kbNameRel;          // unique kb name for the "name" relation

  /* built-in predicates */
  public static String P_INSTANCE       = "@instance";
  public static String P_ISA            = "@isa";
  public static String P_EVERY_INSTANCE = "@everyInstance";
  public static String P_STRING_MATCH   = "@stringMatch";
  public static String P_NOCASE_STRING_MATCH = "@stringIMatch";

  public static String P_EQUAL           = "@=";
  public static String P_NOT_EQUAL       = "@!=";
  public static String P_GREATER         = "@>";
  public static String P_GREATER_OR_EQUAL= "@>=";
  public static String P_LESS_OR_EQUAL   = "@<=";
  public static String P_LESS            = "@<";

  /** Constructs a new SHOE KB interface using the specified API and
   ontology manager. Any messages reported by this object will be
   written to the specified log. */


  public ShoeKb(KBInterface kb, OntManager ontman, PrintStream logOut) {

    this.kb = kb;
    this.ontman = ontman;
    this.logOut = logOut;
    this.implicitClaims = false;
  }

  /** Constructs a new SHOE KB interface using the specified API and
   ontology manager. Any messages reported by this object will be
   written to the specified log. */
  public ShoeKb(KBInterface kb, OntManager ontman, PrintStream logOut,
		boolean implicitClaims) {

    this.kb = kb;
    this.ontman = ontman;
    this.logOut = logOut;
    this.implicitClaims = implicitClaims;
  }


  /** Establish a connection to the given host and port number. */
  public void establishConnection(String host, int port)
       throws KBException
  {
    kb.establishConnection(host, port);
  }

  /** Creates a new KB with the name supplied by the user. The
   necessary meta structures for cataloging ontology information are
   also created. */
  public void createKB(String kbName)
       throws KBException
    {
      kb.createKB(kbName);

      // create necessary structure in KB
      kb.createClass("Ontology");
      kb.createPredicate("_ontId", 2);
      kb.createPredicate("_ontVer", 2);
      kb.createPredicate("_ontURL", 2);
      kb.createPredicate("_ontSeq", 2);

      storedOnts = new Hashtable();
      storeBaseOntology();
    }


  /** Using a KBMS that was specified on construction of the object,
    opens the given knowledge base using a connection specified by the
    given host and port number. */
  public void openKB(String kbName, String host, int port)
       throws KBException
  {
    try {
      kb.openKB(kbName, host, port);
    } catch (KBException ex) {
      storedOnts = new Hashtable();
      throw ex;
    }

    // determine what onts are in the kb
    storedOnts = getStoredOntologies();
    // storeBaseOntology();
  }

  /** Using a KBMS that was specified on construction of the object,
    opens the given knowledge base using the existing connection,
    if one exists. */
  public void openKB(String kbName)
       throws KBException
  {
    try {
      kb.openKB(kbName);
    } catch (KBException ex) {
      storedOnts = new Hashtable();
      throw ex;
    }

    // determine what onts are in the kb
    storedOnts = getStoredOntologies();
    // storeBaseOntology();
  }

  /** Closes the current knowledge base. */
  public void closeKB() {

    kb.closeKB();
  }

  /** Closes the current connection. Any open KBs will be closed as well. */
  public void closeConnection() {

    kb.closeConnection();
  }


  /** Stores the SHOE information from the given document in the KB.
   Any claims originally stored for the source are deleted, and any
   new ontologies are loaded before the new claims are stored. */
  public void storeSHOEDoc(SHOE_Doc doc, String sourceUrl) {

    Enumeration enum;
    SHOE_Obj docContent;
    Instance inst;
    Date parseDate;

    // urlStr = vnode.url.toString();

    docMessages = "";

    /* delete old assertions for this page before storing new ones. Also
     delete claims that were implied by the source (if any). */
    // System.out.println("Deleting existing claims for source");
    kb.deleteClaims(sourceUrl);
    kb.deleteClaims(sourceUrl + "#implied");
    // System.out.println("Deleted existing claims");

    /* examine claims made within the document */
    if (doc.getVersion() == 1.0) {
      doc.validate(ontman);
      enum = doc.getPSHOE();
      while (enum.hasMoreElements()){
	docContent = (SHOE_Obj)enum.nextElement();
	if (docContent instanceof Instance) {
	  inst = (Instance)docContent;
	  // storeInstance() will will first store any requisite ontologies
	  storeInstance(inst, sourceUrl);
	}
      }
    }
  }

  /** Given an instance and the URL of the page on which it resides,
     stores the claims made by the page in the KB. If the implicitClaims
     variable is set, it also makes any claims implied by the instance.
     If the instance requires any ontologies that are not currently
     cataloged in the KB, it will store these ontologies in the KB first. */
  /* BUG: Currently, we use the sourceUrl as the claimant. However,
     the SHOE spec considers the instance key as the claimant.
     The problem with that approach is then we have no way of
     knowing which document the claim came from and thus could not
     properly determine which claims to remove when a document is
     re-processed. Eventually, we need an extra predicate to indicate
     the source document of the claim, so that we can use the claims()
     predicate as it was intended. */
  public void storeInstance(Instance inst, String sourceUrl) {

    Enumeration enum;
    SHOE_Obj obj;
    Relation rel;
    Category cat;

    /*
    sourceUrl = inst.getKey();     // temp override of source (see NOTE above)
    */

    // store any requisite ontologies that are unknown to the KB
    storeNewOntologies(inst.getUseOnts());

    Hashtable impCats = new Hashtable();

    enum = inst.getDeclarations();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Relation) {
	rel = (Relation)obj;
	storeRelation(rel, sourceUrl, impCats);
      }
      else if (obj instanceof Category) {
	cat = (Category)obj;
	storeCategory(cat, sourceUrl);
      }
    }

    /* NOTE: since we make the implicit claims on an instance by instance
       basis instead of a document basis, we may make the same implicit
       claim many times for the same document, up to once per each
       instance in the document. */
    if (implicitClaims)
      makeImplicitClaims(sourceUrl, inst, impCats);
  }


  /** Given an array of Use-Ontologies, determines which of the
    referenced ontologies are not currently loaded into the KB, and
    stores them into it.
    */
  public void storeNewOntologies(Use_Ontology[] useonts) {

    String key;
    Use_Ontology uont;
    Ontology ont;

    for (int i=0; i < useonts.length; i++) {
      uont = useonts[i];
      key = Ontology.makeKey(uont.getId(), uont.getVersion());
      /* if this ontology is not in our hashtable, store it in the KB and
	 add it to the table */
      if (storedOnts.containsKey(key) == false) {
	ont = ontman.getOntology(uont);
	if (ont != null) {
	  /* storeOntology() will add a catalog entry for the ontology */
	  storeOntology(ont);
	}
      }
    }
  }


  /** Stores the ontology's DEF_CATEGORYs, DEF_RELATIONs and DEF_CONSTANTs
    in the KB. For DEF_CATEGORYs, the isa links and names are stored.
    For DEF_RELATIONs a new predicate of the correct arity is created.
    For DEF_CONSTANTs, only the instance links are stored. Stores
    catalog information about the ontology as well, including a
    sequence number assigned to it. */
  public void storeOntology(Ontology ont) {

    SHOE_Obj obj;
    Def_Category dcat;
    Def_Relation drel;
    Def_Constant dconst;
    Def_Inference dinf;
    Vector   rules;
    Enumeration enum;

    // store any requisite ontologies that are unknown to the KB
    /* By calling storeNewOntologies before assigning a sequence
       number, the sequence numbers should be assigned first
       to those ontologies on which an ontology depends. */
    storeNewOntologies(ont.getUseOnts());

    docMessages = docMessages + "Storing ontology " + ont.getOntLabel() + "\n";

    addToOntCatalog(ont);

    setBackwardCompatibility(ont);
    setExtendedOnts(ont);

    enum = ont.getElements();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Def_Category) {
	dcat = (Def_Category)obj;
	storeDefCategory(dcat, ont);
      }
      else if (obj instanceof Def_Relation) {    	// create predicates
	drel = (Def_Relation)obj;
	storeDefRelation(drel, ont);
      }
      else if (obj instanceof Def_Constant) {
	dconst = (Def_Constant)obj;
	storeDefConstant(dconst, ont);
      }
    }

    rules = ont.getInferenceRules();
    enum = rules.elements();
    while (enum.hasMoreElements()) {
      dinf = (Def_Inference)enum.nextElement();
      storeInferenceRule(dinf, ont);
    }
  }


  /** Sets the ontology against which queries will be issued unless
    the query explicitly specifies an ontology. */
  public void setDefaultOntology(String ontId, String ontVer) {

    defaultOntId = ontId;
    defaultOntVer = ontVer;
  }

  /* NOTE: Somehow, the three issueQuery() methods below must be able
     to distinguish between built-in and ontology predicates, so
     that they don't try to translate them to unique ontology names. */

  /** Issues a single atom, n-ary query to the KB using the default
    ontology. Variable arguments are preceded by question marks. The
    answers are returned in a matrix. */
  public ResultSet issueQuery(String predicate, String[] args) {

    Query     qry;
    ResultSet rset;

    qry = new Query(predicate, args);
    rset = issueQuery(qry);
    return rset;
  }

  /** Issues a single atom, binary query to the KB using the default
    ontology. Variables arguments are preceded by question marks. The
    answers are returned in a matrix. */
  public ResultSet issueQuery(String predicate, String arg1, String arg2) {

    Query     qry;
    ResultSet rset;

    qry = new Query(predicate, arg1, arg2);
    rset = issueQuery(qry);
    return rset;
  }

  /** Issues an arbitrary query to the KB using the default
    ontology. Variables arguments are preceded by question marks. The
    answers are returned in a matrix. */
  public ResultSet issueQuery(Query query) {

    Query kbQuery;
    ResultSet rset;

    kbQuery = toKbQuery(query);
    rset = kb.queryKB(kbQuery);
    return rset;
  }


  /** Returns an array consisting of the OntDescriptors of every
    ontology known to the KB. */
  /* NOTE: consider returning the ontologies in a particular order,
     e.g., order of discovery, alpahabetical, etc. */
  public OntDescriptor[] getKnownOntologies() {

    OntDescriptor[] knownOnts;
    Enumeration     enum;
    OntDescriptor   ontDesc;
    int             i = 0;

    knownOnts = new OntDescriptor[storedOnts.size()];
    enum = storedOnts.elements();
    while (enum.hasMoreElements()) {
      ontDesc = (OntDescriptor)enum.nextElement();
      knownOnts[i] = ontDesc;
      i++;
    }
    return knownOnts;
  }


  /** Returns the messages generated during the course of storing the
    current document. */
  public String getDocMessages() {

    return docMessages;
  }


  /* ********************* Protected Methods ************************* */

  protected void addToOntCatalog(Ontology ont)
  {
    String key;                  // ontology's key
    OntDescriptor ontDesc;

    key = ont.getKey();
    maxSeqNum = maxSeqNum + 1;
    ontDesc = new OntDescriptor(ont.getId(), ont.getVersion(), ont.getURL(),
				maxSeqNum);

    /* update stored onts for reference later in this session */
    storedOnts.put(key, ontDesc);

    /* update the KB's catalog of ontologies */
    kb.storeClaim("System", "_ontId", key, ont.getId());
    kb.storeClaim("System", "_ontVer", key, ont.getVersion());
    kb.storeClaim("System", "_ontSeq", key, String.valueOf(maxSeqNum));
    if (ont.getURL() != null)
      kb.storeClaim("System", "_ontURL", key, ont.getURL());
    else
      // if we don't store something then, getStoredOntologies won't
      // retrieve the ontology!
      kb.storeClaim("System", "_ontURL", key, "-");
    kb.storeClaim("System", KBInterface.P_INSTANCE, "Ontology", key);
  }

  protected int getOntSeqNumber(String ontId, String ontVer) {

    OntDescriptor ontDesc;

    ontDesc = getOntDescriptor(ontId, ontVer);
    if (ontDesc == null) {
      System.out.println("ERROR: ShoeKb.getOntSeqNumber(): " +
			 "Couldn't find Ont Descriptor for " +
			 ontId + ",v." + ontVer);
      return -1;
    }
    else
      return ontDesc.getSeqNum();
  }

  protected String getOntURL(String ontId, String ontVer) {

    OntDescriptor ontDesc;

    ontDesc = getOntDescriptor(ontId, ontVer);
    if (ontDesc == null)
      return null;
    else
      return ontDesc.getUrl();
  }

  protected OntDescriptor getOntDescriptor(String ontId, String ontVer) {

    String key;

    key = Ontology.makeKey(ontId, ontVer);
    return (OntDescriptor)storedOnts.get(key);
  }


  /** Retrieves the catalog information for ontologies from the KB. */
  protected Hashtable getStoredOntologies() {

    Hashtable storedOnts;
    Query query;
    ResultSet rset;
    int i;
    int seqNum;
    String ontUrl;
    OntDescriptor ontDesc;

    storedOnts = new Hashtable();
    // the ordering of these conjuncts is important to get the right
    // bindings in the expected places!
    query = new Query("instance", "Ontology", "?X");
    query.addConjunct("_ontSeq", "?X", "?S");
    query.addConjunct("_ontURL", "?X", "?U");
    query.addConjunct("_ontId", "?X", "?I");
    query.addConjunct("_ontVer", "?X", "?V");
    rset = kb.queryKB(query);
    if (rset != null) {
      for (i=0; i < rset.getNumResults(); i++) {
	// System.out.println("Answers[" + i + "][0] = " + answers[i][0]);
	try {
	  seqNum = Integer.parseInt(rset.getValue(i,1));
	} catch (NumberFormatException ex) {
	  System.out.println("Ontology sequence number " + rset.getValue(i,1) +
			     " is not an integer.");
	  return storedOnts;
	}
	ontUrl = rset.getValue(i,2);
	if (ontUrl.equals("-"))
	  ontUrl = null;
	ontDesc = new OntDescriptor(rset.getValue(i,3), rset.getValue(i,4),
				    ontUrl, seqNum);
	storedOnts.put(rset.getValue(i,0), ontDesc);
	if (seqNum > maxSeqNum)
	  maxSeqNum = seqNum;
      }
    }
    return storedOnts;
  }

  /** Old version that only retrieve the key of ontologies known to
    the KB. */
  /*
  protected Hashtable getStoredOntologies() {

    Hashtable storedOnts;
    ResultSet rset;
    int i;

    storedOnts = new Hashtable();
    // we don't use the local P_INSTANCE below because this is direct
    // query to the underlying knowledge base.
    rset = kb.queryKB(KBInterface.P_INSTANCE, "Ontology", "?X");
    if (rset != null) {
      for (i=0; i < rset.getNumResults(); i++) {
	// System.out.println("Answers[" + i + "][0] = " + answers[i][0]);
	storedOnts.put(rset.getValue(i,0), "");
      }
    }
    return storedOnts;
  }
  */


  /** Given the Unique Element identifier for an ontology element,
    constructs a unique name string for it by appending an
    ontology unique sequence number. */
  protected String getUniqueName(Unique_Elt id) {

    String uName = null;
    int seqNum;

    if (id != null) {
      seqNum = getOntSeqNumber(id.getOntId(), id.getOntVer());
      uName = id.getName() + "." + seqNum;
    }
    return uName;
  }


  /** Given the prefixed name of an ontology object, and an ontology with which
    to reference the name, creates and returns a unique name for it.
    The name is constructed by appending an ontology unique sequence
    number. */
  protected String getUniqueName(Ontology refOnt, String name) {

    String uName = null;
    Unique_Elt rootElmtId;

    if (name.startsWith("@"))
      uName = name.substring(1);
    else {
      if (refOnt == null)
	return null;
      try {
	rootElmtId = ontman.resolveName(name, refOnt);
      } catch (SHOEException ex) {
	System.err.println(ex.getMessage());
	return null;
      }
      uName = getUniqueName(rootElmtId);
    }
    return uName;
  }


  /** Given the prefixed name of an ontology object, and the id of an
    ontology with which to reference the name, creates and returns a
    unique name for it. The name is constructed by appending an
    ontology unique sequence number. */
  protected String getUniqueName(String refOntId, String refOntVer,
				 String name)
  {
    String uName = null;
    String refUrl;
    Ontology refOnt;

    refUrl = getOntURL(refOntId, refOntVer);
    if (refUrl != null)
      // if we have a URL for the ontology, provide it to the OntManager
      refOnt = ontman.getOntology(refOntId, refOntVer, refUrl);
    else
      // otherwise, hope that the OntManager can find the ontology anyway
      refOnt = ontman.getOntology(refOntId, refOntVer);
    if (refOnt != null)
      uName = getUniqueName(refOnt, name);
    return uName;
  }


  /* Old version of getUniqueName() that doesn't append a sequence
     number. */
  /*
  protected String getUniqueName(String ontId, String ontVer, String name) {

    String uName = null;

    if (name.startsWith("@"))
      uName = name.substring(1);
    else
      uName = name;
    return uName;
  }
  */


  /** Update the KB to make the given ontology backwards compatible
    with any appropriate prior versions. By default, backward-compatibility
    is ignored. */
  protected void setBackwardCompatibility(Ontology ont)
  {
  }

  /** Update the KB with information about ontologies extended by
    the given ontology. In a context based system, this can be used
    to provide axioms that relate contexts. */
  protected void setExtendedOnts(Ontology ont)
  {
  }


  /** Given a SHOE Category, Relation or Comparison, returns the
    equivalent Atom. */
  protected Atom shoeObjToAtom(SHOE_Obj obj) {

    if (obj instanceof Relation)
      return relationToAtom((Relation)obj);
    else if (obj instanceof Category)
      return categoryToAtom((Category)obj);
    else if (obj instanceof Comparison)
      return comparisonToAtom((Comparison)obj);
    else {
      System.out.println("Bad input to shoeObjToAtom(): " + obj.toString());
      return null;
    }
  }


  /** Given a SHOE category and the URL of the page on which it resides,
    stores it in the knowledge base. */
  protected void storeCategory(Category cat, String sourceUrl) {

    String uniqueCatName;

    if (cat.getElmtId() != null) {
      uniqueCatName = getUniqueName(cat.getElmtId());
      kb.storeClaim(sourceUrl, KBInterface.P_INSTANCE, uniqueCatName,
		    cat.getForText());
    }
  }


  /** Given a SHOE DEF_CATEGORY and its source ontology, stores it in
    the KB. The isa links and names are stored. */
  protected void storeDefCategory(Def_Category dcat, Ontology ont) {

    String objKbName;               // the KB specific name of a SHOE object
    Enumeration pElmtEnum;
    Unique_Elt elmt_id;
    String parKbName;             // the KB specific name of a parent category
    String key;

    key = ont.getKey();

    // System.out.println("Storing category " + dcat.getName());
    objKbName = getUniqueName(ont, dcat.getName());

    kb.createClass(objKbName);

    // store short name of category
    // disable the following condition to run OKBC version (can't use name
    // relation until it is loaded from the base ontology)
    if (dcat.getPhrase() != null)
      kb.storeClaim(key, kbNameRel, objKbName, dcat.getPhrase());

    // store structural links to parents
    pElmtEnum = dcat.getParentIds();
    while (pElmtEnum.hasMoreElements()) {
      elmt_id = (Unique_Elt)pElmtEnum.nextElement();
      parKbName = getUniqueName(elmt_id);
      // store structural link
      kb.storeClaim(key, KBInterface.P_ISA, objKbName, parKbName);
    }
  }


  /** Given a SHOE DEF_CONSTANT and its source ontology, stores it in
    the KB. Only the instance links are stored. */
  protected void storeDefConstant(Def_Constant dconst, Ontology ont) {

    String objKbName;               // the KB specific name of a SHOE object
    String parKbName;             // the KB specific name of a parent category
    Unique_Elt elmt_id;
    String key;

    key = ont.getKey();

    objKbName = getUniqueName(ont, dconst.getName());
    elmt_id = dconst.getCategoryId();
    if (elmt_id != null) {
      parKbName = getUniqueName(elmt_id);
      kb.storeClaim(key, KBInterface.P_INSTANCE, parKbName, objKbName);

      // must decide if we want to add a SHORT field to DEF-CONSTANT
      /*
	// store short name of constant
	  if (dconst.getPhrase() != null)
	    kb.storeClaim(key, kbNameRel, objKbName, dconst.getPhrase());
       */
    }
  }


  /** Given a SHOE DEF_RELATION and its source ontology, stores it in
    the KB. A new predicate of the correct arity is created. */
  protected void storeDefRelation(Def_Relation drel, Ontology ont) {

    String objKbName;               // the KB specific name of a SHOE object
    String key;
    Def_Arg[] dargs;
    String[] argTypes;
    Unique_Elt argTypeId;

    key = ont.getKey();

    objKbName = getUniqueName(ont, drel.getName());

    // kb.createPredicate(objKbName, drel.getArity());

    dargs = drel.getDefArgs();
    argTypes = new String[dargs.length];
    for (int i=0; i<dargs.length; i++) {
      argTypeId = dargs[i].getTypeId();
      argTypes[i] = getUniqueName(argTypeId);
    }
    kb.createPredicate(objKbName, argTypes);

    /* NOTE: Add code for infering types from relation arguments here! */
  }


  /** Given a SHOE DEF-INFERENCE rule and its source ontology, asserts it
    the KB. */
  protected void storeInferenceRule(Def_Inference dinf, Ontology ont) {

    Axiom rule;
    Enumeration ants;
    Enumeration cons;
    SHOE_Obj obj;
    Atom conjunct;

    // System.out.println("Storing DEF-INFERENCE: " + dinf.toString());
    rule = new Axiom();
    ants = dinf.getInfIf().elements();
    while (ants.hasMoreElements()) {
      obj = (SHOE_Obj)ants.nextElement();
      conjunct = shoeObjToAtom(obj);
      if (conjunct != null)
	rule.addAntecedent(conjunct);
      else
	System.out.println(obj.toString() + " resulted in a null conjunct");
    }
    cons = dinf.getInfThen().elements();
    while (cons.hasMoreElements()) {
      obj = (SHOE_Obj)cons.nextElement();
      conjunct = shoeObjToAtom(obj);
      if (conjunct != null)
	rule.addConsequent(conjunct);
      else
	System.out.println(obj.toString() + " resulted in a null conjunct");
    }
    kb.storeAxiom(rule);
  }

  /** Given a SHOE relation and the URL of the page on which it resides,
    stores it in the knowledge base. If the implicitClaims variable is
    set, it also tracks any claims implied by the relation. */
  protected void storeRelation(Relation rel, String sourceUrl,
			       Hashtable impCats)
  {
    RelArg argList[];
    String argVals[];
    int i;
    Unique_Elt ontRef;
    String predName;

    if (rel.getElmtId() != null) {        // only store valid relations
      argList = rel.getArgs();

      /* Get an array of argument values. If the value is a reference to
	 an ontology object, use the unique name of the ontology object */
      argVals = new String[argList.length];
      for (i=0; i < argList.length; i++) {
	ontRef = argList[i].getOntReference();
	if (ontRef != null)
	  argVals[i] = getUniqueName(ontRef);
	else
	  argVals[i] = argList[i].getValue();
      }

      predName = getUniqueName(rel.getElmtId());
      /* Note: storeClaim will escape characters if necessary */
      kb.storeClaim(sourceUrl, predName, argVals);
      // kb.storeClaim(sourceUrl, predName, argVals[0], argVals[1]);

      /* if implicitClaims is set, remember any category claims
	 implied by the relation */
      if (implicitClaims)
	setImpliedCategories(impCats, rel);
    }
  }


  /** Given a generic Atom, transforms it into an Atom suitable for
    use with the KB. Transforms category names that appear in the
    strucutural predicates to their internal KB versions. */
  /* BUG: What about translating constants to unique ontology names?
     How can we tell the difference between a SHOE constant and a
     string? */
  protected Atom toKbAtom(Atom atom) {

    String   kbPred;
    String[] args;
    String[] kbArgs;
    Atom     kbAtom;
    int      i;

    kbPred = getUniqueName(defaultOntId, defaultOntVer, atom.getPredicate());
    args = atom.getArguments();
    kbArgs = new String[args.length];
    for (i=0; i < args.length; i++)
      kbArgs[i] = args[i];
    if (kbPred.equals(KBInterface.P_EVERY_INSTANCE))
      kbArgs[0] = getUniqueName(defaultOntId, defaultOntVer, kbArgs[0]);
    else if (kbPred.equals(KBInterface.P_INSTANCE))
      kbArgs[0] = getUniqueName(defaultOntId, defaultOntVer, kbArgs[0]);
    else if (kbPred.equals(KBInterface.P_ISA)) {
      kbArgs[0] = getUniqueName(defaultOntId, defaultOntVer, kbArgs[0]);
      kbArgs[1] = getUniqueName(defaultOntId, defaultOntVer, kbArgs[1]);
    }
    kbAtom = new Atom(kbPred, kbArgs);
    return kbAtom;
  }


  /* Translate the query into a version that uses KB names for objects
     relative to the default ontology. */
  protected Query toKbQuery(Query query)
  {
    Query kbQuery;
    Enumeration atomEnum;
    Atom atom;
    Atom kbAtom;

    kbQuery = new Query();
    kbQuery.setNumAnswersRequested(query.getNumAnswersRequested());
    atomEnum = query.getAtoms().elements();
    while (atomEnum.hasMoreElements()) {
      atom = (Atom)atomEnum.nextElement();
      kbAtom = toKbAtom(atom);
      kbQuery.addConjunct(kbAtom);
    }
    return kbQuery;
  }


  /* ********************* Private Methods ************************* */


  /** Given a SHOE Category, returns the equivalent Atom. */
  private Atom categoryToAtom(Category cat) {

    Atom atom = null;
    String uniqueCatName;
    String inst;                    // id for instance of the category

    if (cat.getElmtId() != null) {
      uniqueCatName = getUniqueName(cat.getElmtId());
      inst = cat.getForText();
      if (cat.isVariable())
	inst = "?" + inst;
      atom = new Atom(KBInterface.P_INSTANCE, uniqueCatName, inst);
    }
    return atom;
  }


  /** Given a SHOE comparison, returns the equivalent Atom. */
  private Atom comparisonToAtom(Comparison comp) {

    Atom   atom = null;
    RelArg argList[];
    String argVals[];
    int i;
    Unique_Elt ontRef;
    String predName = "";
    int compOp;

    argList = comp.getArgs();

    /* Get an array of argument values. If the value is a reference to
       an ontology object, use the unique name of the ontology object */
    argVals = new String[argList.length];
    for (i=0; i < argList.length; i++) {
      ontRef = argList[i].getOntReference();
      if (ontRef != null)
	argVals[i] = getUniqueName(ontRef);
      else
	argVals[i] = argList[i].getValue();

      // add a question mark in front of variables
      if (argList[i].isVariable())
	argVals[i] = "?" + argVals[i];
    }

    compOp = comp.getOperator();
    switch (compOp) {
    case SHOE_Obj.C_EQUAL:
      predName = P_EQUAL;
      break;
    case SHOE_Obj.C_NOT_EQUAL:
      predName = P_NOT_EQUAL;
      break;
    case SHOE_Obj.C_GREATER:
      predName = P_GREATER;
      break;
    case SHOE_Obj.C_GREATER_OR_EQUAL:
      predName = P_GREATER_OR_EQUAL;
      break;
    case SHOE_Obj.C_LESS_OR_EQUAL:
      predName = P_LESS_OR_EQUAL;
      break;
    case SHOE_Obj.C_LESS:
      predName = P_LESS;
    }
    atom = new Atom(predName, argVals);
    return atom;
  }


  /** Given a SHOE relation, returns the equivalent Atom. */
  private Atom relationToAtom(Relation rel) {

    Atom   atom = null;
    RelArg argList[];
    String argVals[];
    int i;
    Unique_Elt ontRef;
    String predName;

    if (rel.getElmtId() != null) {        // only convert valid relations
      argList = rel.getArgs();

      /* Get an array of argument values. If the value is a reference to
	 an ontology object, use the unique name of the ontology object */
      argVals = new String[argList.length];
      for (i=0; i < argList.length; i++) {
	ontRef = argList[i].getOntReference();
	if (ontRef != null)
	  argVals[i] = getUniqueName(ontRef);
	else
	  argVals[i] = argList[i].getValue();

	// add a question mark in front of variables
	if (argList[i].isVariable())
	  argVals[i] = "?" + argVals[i];
      }

      predName = getUniqueName(rel.getElmtId());
      /* Note: storeClaim will escape characters if necessary */
      atom = new Atom(predName, argVals);
    }
    return atom;
  }


  /** If needed, updates the catalog to reflect the existence of the
    base ontology and stores it in the KB. Also initializes the
    kbNameRel which is the system name of the "name" relation. */
  private void storeBaseOntology() {

    String   key;
    Ontology baseOnt;

    key = Ontology.makeKey(BASE_ONT_ID, BASE_ONT_VER);
    if (storedOnts.containsKey(key) == false) {
      System.out.println("StoredOnts = " + storedOnts.toString());
      System.out.println("Storing the base ontology in the KB");
      baseOnt = ontman.getOntology(BASE_ONT_ID, BASE_ONT_VER);
      storeOntology(baseOnt);
    }

    /* Compute the KB version of the "name" relation. This name
       depends on the base ontology being in the storedOnts (the
       catalog of ontologies). Therefore, this cannot occur in the
       constructor because storedOnts is not set until a KB is
       opened. */
    kbNameRel = getUniqueName(BASE_ONT_ID, BASE_ONT_VER, "name");

  }


  /* The next two methods are only used in non-axiomatic KBs that
     implement the semantics of typed relation arguments, i.e.,
     an instance that appears in a typed argument position can be
     inferred to be a member of that class. */

  /** Explicitly stores any claims that are implied by the specified
    instance, but cannot be inferred from its category claims. The
    candidate categories are given in the form of the impCats hashtable.
    Note: We must be careful to not make the same implict claims over
     and over again as a result of different relations. */
  private void makeImplicitClaims(String sourceUrl, Instance inst,
				  Hashtable impCats) {

    ClaimSubjectIndexer claimIndex;
    Enumeration keys;
    String      key;
    Hashtable   subCats;
    Enumeration impCatEnum;
    Vector      impCatVec;
    Unique_Elt  catId;
    Vector      catClaims;
    int         i, j;
    Unique_Elt  candCat;
    Enumeration explicitEnum;
    boolean     redundant;
    Category    explicitCat;
    Unique_Elt  explicitId;
    Unique_Elt  cat1, cat2;
    Enumeration newCatEnum;
    Unique_Elt  newCatId;
    String      uniqueCatName;

    claimIndex = new ClaimSubjectIndexer(inst, 0);
    keys = impCats.keys();
    while (keys.hasMoreElements()) {
      key = (String)keys.nextElement();

      /* create a vector of category ids implied for this subject */
      subCats = (Hashtable)impCats.get(key);
      impCatEnum = subCats.keys();
      impCatVec = new Vector();
      while (impCatEnum.hasMoreElements()) {
	catId = (Unique_Elt)impCatEnum.nextElement();
	impCatVec.addElement(catId);
      }

      catClaims = claimIndex.getSubjectCategories(key);

      /* Remove any candidate categories that are implied by existing
	 category claims */
      for (i=impCatVec.size()-1; i >= 0; i--) {
	candCat = (Unique_Elt)impCatVec.elementAt(i);
	explicitEnum = catClaims.elements();
	redundant = false;
	while (explicitEnum.hasMoreElements() && redundant == false) {
	  explicitCat = (Category)explicitEnum.nextElement();
	  explicitId = explicitCat.getElmtId();
	  if (ontman.isSubcategoryOf(explicitId, candCat))
	    redundant = true;
	}
	if (redundant == true)
	  impCatVec.removeElementAt(i);
      }

      /* Remove any candidate categories that are implied by another
	 candidate category. */
      for (i=impCatVec.size()-1; i >= 0; i--) {
	redundant = false;
	for (j=0; j < impCatVec.size() && redundant == false; j++) {
	  if (i != j) {
	    cat1 = (Unique_Elt)impCatVec.elementAt(i);
	    cat2 = (Unique_Elt)impCatVec.elementAt(j);
	    if (ontman.isSubcategoryOf(cat2, cat1))
	      redundant = true;
	  }
	}
	if (redundant == true)
	  impCatVec.removeElementAt(i);
      }


      /* Add claims for each non-redundant implied category. */
      newCatEnum = impCatVec.elements();
      while (newCatEnum.hasMoreElements()) {
	newCatId = (Unique_Elt)newCatEnum.nextElement();
	uniqueCatName = getUniqueName(newCatId);
	kb.storeClaim(sourceUrl +"#implied", KBInterface.P_INSTANCE,
		      uniqueCatName, key);
      }
    }
  }

  /* Alternate aproaches to implict claims:

     1) iterate for each argument, create a table of children for
        that argument and then cycle through all claims to determine
	if any say that the instance in question is a member of one
	of these children.
     2) create a table of parent category declarations and parents
        for each subject described by this instance. for each
	argument in which the subject appears, check if the
	associated table contains the correct type
  */


  /** Given a hash table which stored implied categories for specific
    instance keys, adds all of the categories implied by the specified
    relation. A category claim is implied when the instance uses a
    particular key in an argument position that requires an object of
    a certain category. Note that the contents of the hashtable are
    hashtables themselves, and their keys are the category names. This
    makes it possible to qucikly determine if a particular category has
    been added for a particular instance key. */
  private void setImpliedCategories(Hashtable impCats, Relation rel) {

    Unique_Elt    defId;
    Def_Relation  defRel;
    RelArg[]      args;
    RelArg        arg;
    int           pos;
    int           i;
    Def_Arg       defArg;
    Unique_Elt    typeId;
    String        key;
    Hashtable     subCats;

    defId = rel.getElmtId();
    defRel = (Def_Relation)ontman.getElement(defId);

    args = rel.getArgs();
    for (i=0; i < args.length; i++) {
      arg = args[i];
      pos = arg.getPosition();
      defArg = defRel.getDefArg(pos);
      if (defArg != null) {
	if (defArg.getTypeClass() == Def_Arg.CATEGORY) {
	  typeId = defArg.getTypeId();
	  if (typeId != null) {
	    key = arg.getValue();
	    subCats = (Hashtable)impCats.get(key);
	    if (subCats == null) {
	      subCats = new Hashtable();
	      subCats.put(typeId, "");
	      impCats.put(key, subCats);
	    }
	    else
	      subCats.put(typeId, "");
	  }
	}
      }
    }
  }

  /**below is added by zpan*/
  public String getdefaultOntId()
  {
  return this.defaultOntId;
  }

  public String getdefaultOntVer()
  {
  return this.defaultOntVer;
  }

  protected void setdocMessages(String message)
  {
  docMessages=message;
  }
}
