/* SHOE Search
 * Copyright (C) 1999 PLUS Group - University of Maryland
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Authors: Jeff Heflin and Carolyn Gasarch
 * Contact: Jeff Heflin (heflin@cs.umd.edu)
 *
 */

/* Change History:
   12/29/00 JDH - Changed initAvailOnts() to work with Ontology.getKnownOnts()
   5/30/02  Zpan - Changed getKbInterface() to abstract, ShoeParka.SS_Frame will
   implement it.
*/

package ShoeSearch;

import SHOE.*;
import ShoeKb.*;
import Tree.*;
import UserInt.ApplType;
import UserInt.ArrayMenuBar;
import UserInt.BorderedPanel;
import UserInt.GridBagCreator;
import UserInt.LightComponentScroller;
import UserInt.MultiColumn;
import UserInt.Notification;
import UserInt.TreeWidget;
import java.applet.AppletContext;
import java.applet.Applet;
import java.awt.*;
import java.util.*;
import java.net.*;
import java.io.*;

/** This is the main class of the ShoeSearch application. It sets up the
  various panels and responds to user events.
*/
//getKbInterface() method is abstract.
public abstract class SS_Frame extends Frame implements WebSearchContainer
{
  protected ShoeKb      kb;
  protected OntManager  ontman;            // ontology cache
  protected Ontology    contextOnt;        // context ontology

  /** True if properties with no values should not be allowed to be queried */
  protected boolean     disableEmptyProps = false;

  protected QueryPanel  qwin;         // panel that allows user to make query
  protected ResultsWin  resWin;       // panel that displays results of query
  LightComponentScroller sqwin;       // component that scrolls qwin
  BorderedPanel  sqwinPanel;               // panel that contains the scroller

  ArrayMenuBar   menuBar;
  protected Choice         ontChoice;
  protected TreeWidget     catList;
  Button   b_Select;
  Button   b_Query;
  Button   b_Getpage;
  Button   b_WebSearch;
  Button   b_AddToQuery;
  Label    statusBar;

  protected String     kbHost;
  protected int        kbPort;
  protected String     kbName;
  String[][] proxyOnts;
  protected OntDescriptor[] availOnts;

  protected Applet        srcApplet;
  protected AppletContext context;
  WebSearch doWebSearch;

  boolean  correspondence;        // category displayed in QueryPanel and
                                  // responses in ResultsWin (if any)
                                  // correspond; 'false' means a diff category
                                  // is now in QueryPanel than the query was
                                  // issued for.

  static final String  TITLE = "SHOE Search";
  static final String  URL_PREFIX   = "http:";
  static final String  WEBPG_TARGET = "SHOE Document";  //separate window

  /* Menu item constants */
  static final String  MI_FILE = "File";
  static final String  MI_FWD_WS = "WebSearch Window";
  static final String  MI_FWD_WSDEP = "WebSearchHelp Window";
              //dependent window to "WebSearch Window"
  static final String  MI_EXIT = "Exit";
  static final String  MI_HELP = "Help";

  /* Default server settings  */
  /* These can be set to custom values by subclassing this class. */
  protected static String DEFAULT_HOST    = "";
  protected static int    DEFAULT_PORT    = 0;
  protected static String DEFAULT_KB_NAME = "";

  protected static final String PROXY_FILE = "ontproxy.dat";


    public SS_Frame (String kbHost, int kbPort, String kbName,
		   Applet srcApplet)
  {
      super(TITLE);

      int i;
      int cid = 0;                 // index of default ontology

      /*
      System.out.println("SS_Frame() init with: " + kbHost + ", " +
		       kbPort + ", " + kbName);
      */

      /* Set parameters dealing with the PARKA server and then sets up
         the initial display */

      this.kbHost = kbHost;
      this.kbPort = kbPort;
      this.kbName = kbName;
      this.srcApplet = srcApplet;
      if (srcApplet == null)
	this.context = null;
      else
	this.context = srcApplet.getAppletContext();

      doWebSearch = null;
      correspondence = true;

      /* The OntManager must be created before we set up the components
	 because it is passed to the QueryPanel. */
      ontman = new OntManager();
      setupMenuBar();
      setupComponents(cid);

      /* set proxy onts by reading from file for applications or
	 reading from URL for applets. */
      if (ApplType.getIsApplication())
	loadOntProxyForApplication();
      else
	loadOntProxyForApplet();

      boolean connected;
      connected = connectToKb();
      if (connected == true) {
	initAvailOnts();

	initOntChoiceWidget();
	contextOnt = selectInitialOntology(cid);
      }
      else {
	showStatus("Couldn't connect to KB " + kbName + " at " +
		   kbHost + ":" + kbPort);
      }

      resize (preferredSize());

      /* BUG: The preferred size of sqwin does not take into account the
	 preferred size of its viewport. */
      /*
      System.out.println("Pref sqwin = " + sqwin.getPreferredSize());
      System.out.println("Pref sqwinPanel = " + sqwinPanel.getPreferredSize());
      */

      pack();
      // validate();
  }


  /** When the class is unloaded, make sure that the KB is closed. */
  protected void finalize()
  {
      System.out.println("Finalizing...");
      kb.closeKB();
  }

  /** Opens a SHOE Kb. */
  protected boolean connectToKb() {

    KBInterface baseKb;            // underlying KB for SHOE data
    PrintStream logOut;
    boolean     connected = false;

    baseKb = getKbInterface();
    logOut = System.out;
    kb = getShoeKb(baseKb, ontman, logOut);

    System.out.println("Connecting with parameters: " + kbHost + ", " +
		       kbPort + ", " + kbName);

    try {
      kb.openKB(kbName, kbHost, kbPort);
      connected = true;
    } catch (KBException ex) {
      if (context != null) {
	/* we have not created the frame yet, so send messages to applet
	   context */
	context.showStatus(ex.getMessage());
	logOut.println(ex.getMessage());
	connected = false;
      }
      else {
	logOut.println(ex.getMessage());
	connected = false;
	// should we also beep? prevent queries if no KB was opened?
      }
    }
    return connected;
  }

  void setupMenuBar()
  {
      Object menuItems[][] = {
        {MI_FILE, MI_FWD_WS, MI_FWD_WSDEP, MI_EXIT},
        {MI_HELP},
        };
      menuBar = new ArrayMenuBar(menuItems);
      setMenuBar(menuBar);
  }

  /** Create panels and widgets.
  */
  void setupComponents (int cid)
  {
    Label l = new Label ("Ontology:");
    ontChoice = new Choice();
    BorderedPanel catPanel = new BorderedPanel(3,4,0,4);
    setupCatPanel (catPanel);

    b_Query = new Button("Query");
    qwin = new QueryPanel (ontman);
    sqwin = new LightComponentScroller(qwin);   //"scrolling qwin"
    sqwinPanel = new BorderedPanel(2,4,1,4);
    sqwinPanel.setLayout (new BorderLayout());
    sqwinPanel.add ("Center", sqwin);
    sqwinPanel.setHasBorder();

    resWin = new ResultsWin (this, context);
    // resWin = new ResultsWin (this, kb, context);
    b_Getpage = new Button("Get Page");
    b_WebSearch = new Button("Web Search");
    b_AddToQuery = new Button("Add To Query");
    statusBar = new Label("                                                  ");

    GridBagCreator   gbcreat = new GridBagCreator (this);

    /* addComponent(comp, x, y, width, height, x_wt, y_wt, fill, x_in, y_in) */
    gbcreat.addComponent (l,         0,  0,  1,  1,    0,   0,
                       GridBagConstraints.BOTH, 0, 0);
    gbcreat.addComponent (ontChoice, 1,  0,  2,  1,    0,   0,
                       GridBagConstraints.BOTH, 0, 0);
    gbcreat.addComponent (b_Query,   3,  0,  3,  1,    0,   0,
                       GridBagConstraints.NONE, 0, 0);

    gbcreat.addComponent (catPanel,  0,  1,  3,  1,    20,  60,
                       GridBagConstraints.BOTH, 0, 0);
    gbcreat.addComponent (sqwinPanel, 3,  1,  3,  1,   80,   0,
                       GridBagConstraints.BOTH, 0, 0);
    gbcreat.addComponent (resWin,    0,  2,  GridBagConstraints.REMAINDER, 1,
                       0, 40,
                       GridBagConstraints.BOTH, 0, 0);

    gbcreat.addComponent (b_WebSearch,  0, 3, 1,  1,    0,   0,
                       GridBagConstraints.NONE, 0, 0);
    gbcreat.addComponent (b_Getpage,    1, 3, 2,  1,    0,   0,
                       GridBagConstraints.NONE, 0, 0);
    gbcreat.addComponent (b_AddToQuery, 3, 3, 1,  1,    0,   0,
                       GridBagConstraints.NONE, 0, 0);
    gbcreat.addComponent (statusBar,   4,  3,  2, 1,
                       0, 0, GridBagConstraints.HORIZONTAL, 0, 0);

    setBackground(Color.lightGray);
  }

  protected void initAvailOnts() {

    availOnts = kb.getKnownOntologies();
  }


  /** Fill the values for the choice widget used to select the
    ontology to perform queries against.  */
  public void initOntChoiceWidget ()
  {
    OntDescriptor ontDesc;

    for (int i=0; i < availOnts.length; i++) {
      ontDesc = availOnts[i];
      ontChoice.addItem(ontDesc.getId() + ", v." + ontDesc.getVersion());
    }
  }


  /** Sets the current ontology based on its sequence in the ontology
    pick list. The ontology is set as the default in the KB, the
    choice widget is updated, the query window is notified, and the
    corresponding Ontology object is returned. */
  Ontology selectInitialOntology(int cid)
  {
    Ontology ont = null;

    // use ontChoice.getItemCount() for Java 1.1
    if (ontChoice.countItems() < cid)
      System.err.println("Error: Invalid index for ontology widget.");
    else {
      /* set for use with default ontology */
      String ontURL = availOnts[cid].getUrl();
      // String ontURL = ontServer + availOnts[cid][2];

      //TEST
      System.out.println("Getting " + ontURL);

      ont = ontman.getOntology (availOnts[cid].getId(),
				availOnts[cid].getVersion(), ontURL);
      //TEST
      if (ont != null)
	System.out.println("Got ontology");

      kb.setDefaultOntology(ont.getId(), ont.getVersion());

      ontChoice.select(cid);
      qwin.setCurrentOntology(ont);   // set default ont in QueryWindow
      loadCategoryList (catList, ont);
    }
    return ont;
  }


  /** Set up the panel which contains the category tree.
  */
  public void setupCatPanel (Panel p)
  {
      p.setLayout (new BorderLayout());
      Label l = new Label ("Select a category:");
      catList = new TreeWidget(10, false);
      catList.setToggleClickCount(2);
      b_Select = new Button("Select");
      p.add("North", l);
      p.add("Center", catList);
      p.add("South", b_Select);
  }

  public void showStatus (String status)
  {
      statusBar.setText (status);
  }

  /** Quit the applet or application. */
  public void exit()
  {
    // System.out.println("ShoeSearch.SS_Frame.exit(): Disconnecting from the server");
    kb.closeKB();
    if (ApplType.getIsApplication())
      System.exit(0);
    else {
      System.out.println("Disposing");
      hide();
      dispose();
    }
  }

  /** Implements WebSearchContainer interface. */
  public void closeWebSearchWin()
  {
      doWebSearch.closeWebSearchQHelpWin();
      doWebSearch.dispose();
      doWebSearch = null;
  }

  /** Respond to menu events.
  */
  public boolean handleEvent (Event event)
  {
    if (event.id == Event.WINDOW_DESTROY) {
      exit();
      return true;
    }
    else if (event.target instanceof MenuItem) {
      MenuItem menu = (MenuItem)event.target;
      MenuItem pmenu = (MenuItem)menu.getParent();
      if (event.id == Event.ACTION_EVENT) {
         if (pmenu.getLabel().equals(MI_FILE)) {
            if (MI_EXIT.equals(event.arg))
               exit();
            else  if (MI_FWD_WS.equals(event.arg))
            {
               if (doWebSearch != null)
	               doWebSearch.toFront();
            }
            else  if (MI_FWD_WSDEP.equals(event.arg))
            {
               if (doWebSearch != null)
                  doWebSearch.assocWinToFront();
            }
         }
      }
    }
    else if (event.id == Event.LIST_SELECT) {
      if (event.target == catList) {
         // System.out.println("Selected catList");
      }
    }
    return super.handleEvent(event);
  }

  /** Respond to button presses and choice selections
  */
  public boolean action (Event evt, Object arg)
  {
    if (evt.target instanceof Button)
    {
        if (evt.target == b_Select)
        {
	  if (catList.getSelectedIndex() >= 0) {
	    setCursor(Frame.WAIT_CURSOR);
	    // clear results panel of any answers to previous queries
	    resWin.newEmptyMultiColumn();

	    SHOE_Obj obj = (SHOE_Obj)catList.getSelectedObject();
	    qwin.hide();
	    sqwin.scrollTo(0,0);
	    qwin.changeSubject (contextOnt, obj.getKey());

	    /* We want to make sure the scroll pane will be re-laid out. A
	       call to validate() will not resize/reposition an object
	       and its components unless it has been marked invalid. Note
	       that Java 1.1 Container's invalidate() method overrides
	       that of Component. */
	    if (ApplType.getIsVersion1_0()) {
	      ((Component)sqwin).invalidate();
	      ((Component)sqwinPanel).invalidate();
	    }
	    else {
	      sqwin.invalidate();
	      sqwinPanel.invalidate();
	    }

	    int  qwinSize = qwin.getRowSize();
	    if (disableEmptyProps == true) {
	      for (int i = 0; i < qwinSize; i++)
		{
		  int numResults = doFindQueryForNumAnswers(i);
		  if (numResults == 0)
		    qwin.disableRow(i);
		  //for any fields with no data anyway, their rows are disabled
		}
	    }

	    correspondence = false;

	    //pack();      // causes QueryPanel to be to small


	    qwin.show();

	    // tell sqwinPanel that it can draw its border
	    if (qwinSize > 0)
	      sqwinPanel.setVisibleComponents();

	    validate();
	    repaint();
	    setCursor(Frame.DEFAULT_CURSOR);
	    return true;
	  }
	}
        else if (evt.target == b_Query)
        {             // issue a query command
	  setCursor(Frame.WAIT_CURSOR);
          ResultSet rset = doQuery();   // doQuery() updates ResultWin
          correspondence = true;
	  validate();
	  // repaint();
	  setCursor(Frame.DEFAULT_CURSOR);
          return true;
        }
        else if (evt.target == b_Getpage)
        {
            MultiColumn  mc = resWin.getMultiColumn();
            int    activeRow = mc.selected(),
                   activeCol = mc.selectedCol();
            String text = mc.getItemCell (activeRow, activeCol);
            text.trim();
  /* TEST
            System.out.println ("URL selected:  " + text);
  */
           if (text.startsWith (URL_PREFIX))
           {
              URL  webpage;
              try    {
                 webpage = new URL (text);
                 context.showDocument (webpage, WEBPG_TARGET);
              }
              catch  (MalformedURLException  mue)  {
                 System.out.println ("Malformed URL:  Cannot retreive");
              }
           }

            return true;
        }
        else if (evt.target == b_WebSearch)
        {
            //only create a new win if one of this type does not currently exist
            if (doWebSearch == null)
            {
               doWebSearch = new WebSearch (this,
                           TITLE + " -- Web Search", context, qwin);
                   // doWebSearch.resize (put width here, put height here);
                   //is 'show()' in Window class or Dialog class?
               if (ApplType.getIsVersion1_0())
               {
                   // System.out.println ("Version 1.0");
                   ((Window)doWebSearch).show();
               }
	            else
	            {
                   // System.out.println ("Version 1.1 or higher");
                   doWebSearch.show();
               }
            }
	         else    //put the existing window on top of stack
               doWebSearch.toFront();

            return true;
        }
        else if (evt.target == b_AddToQuery)
        {
            ResultSet rset;
            String    vname;
            int       rnum;

            if (correspondence)
            {
                MultiColumn  mc = resWin.getMultiColumn();
                int    activeRow = mc.selected(),
                       activeCol = mc.selectedCol();

                /* get variable name corresponding to selected column and use
                   it to determine the index of the selected relation */
                rset = resWin.getQR();
                vname = rset.getColVar(activeCol);
                rnum = qwin.getVarRelIndex(vname);

                // get text from cell
                String text = mc.getItemCell (activeRow, activeCol);
                text.trim();

                qwin.setFieldValue(rnum, text);
            }
            else
                System.out.println ("Cannot copy from results unrelated to current Category");
       }
       else
       {
           String label = (String) arg;
           if (label.equals (QueryPanel.FIND_BUTTON_LABEL))
           {
              int  whichButton = qwin.getFindButton ((Button)evt.target);
              if (whichButton != -1)
              {
                 doFindQuery (whichButton);
                 correspondence = true;
              }
              else     //  this is an error cond:
                 System.out.println ("Could not determine which '" +
                        QueryPanel.FIND_BUTTON_LABEL +
                                  "' button was pressed.");
	      validate();
	      repaint();
              return true;
           }
           /* replaced following with the "Add To Query" button:
           else   if (label.equals (QueryPanel.MOVETEXT_BUTTON_LABEL))
           {
              int  whichButton = qwin.getCopyButton ((Button)evt.target);
              if (whichButton != -1)
              {
                 MultiColumn mc = resWin.getMultiColumn();
	              String   selectedText = mc.getItemCell (
                                         mc.selected(), mc.selectedCol());
		           // we want to make sure selectedText is not white space;
                 // but we do not want any changes, else the copied value
                 // will not match what it should match in Parka (Parka
                 // does not ignore leading/trailing whitespace.
                 //
                 String   selCopy = new String (selectedText);
                 selCopy.trim();
                 if (!selCopy.equals (""))
	                 qwin.setFieldValue (whichButton, selectedText);
              }
              else     //  this is an error cond:
                    System.out.println ("Could not determine which '" +
                        QueryPanel.MOVETEXT_BUTTON_LABEL +
                                  "' button was pressed.");

              return true;
           }
           */
        }
    }
    else if (evt.target == ontChoice)
    {
        // System.out.println("Selected ontChoice");
      /* setCursor() may be what causes the applet version to gradually
	 move the frame to the upper left. */
        setCursor(Frame.WAIT_CURSOR);
        changeContext(ontChoice.getSelectedIndex());
	setCursor(Frame.DEFAULT_CURSOR);

        return true;
    }

    return false;
  }

  /** Load an expandable list with the complete category hierarchy
    for the given ontology.
  */
  protected void loadCategoryList (TreeWidget list, Ontology ont) {

    Enumeration catTrees;
    Vector catVec;
    DefaultTreeModel nextTree;

    catVec = new Vector();
    catTrees = ont.getLabeledCategoryTreeLists(ontman);
    while (catTrees.hasMoreElements()) {
      nextTree = (DefaultTreeModel)catTrees.nextElement();
      catVec.addElement(nextTree);
    }
    list.initTree(catVec);
    list.expandAll();
  }

    /** Changes the ontology and knowledge base used by the system to the
      one specified by an index into availOnts.
    */
    public void changeContext (int cid)
    {
        Ontology ont = null;
	String ontId;
	String ontVer;
	String ontURL;

        catList.clear();
        qwin.removeAll();                        // clear query panel
        resWin.newEmptyMultiColumn();           // clear results panel

	ontId = availOnts[cid].getId();
	ontVer = availOnts[cid].getVersion();
        ontURL = availOnts[cid].getUrl();
        // ontURL = ontServer + availOnts[cid][2];

	try {
	  ont = ontman.getOntology(ontId, ontVer, ontURL);
	} catch (Exception ex) {
	  Notification note;
	  String msg = "The ontology " + ontId + ", v." + ontVer +
	    "\ncould not be loaded from " + ontURL + ".\n" +
	    "This may be due to Java applet security restrictions.";
	  note = new Notification(this, "Error", msg);
	  note.showWorkaround();
	  return;
	}
	if (ont == null) {
	  Notification note;
	  String msg = "The ontology " + ontId + ", v." + ontVer +
	    "\ncould not be loaded from " + ontURL + ".";
	  note = new Notification(this, "Error", msg);
	  note.showWorkaround();
	  return;
	}
        qwin.setCurrentOntology (ont);
	kb.setDefaultOntology(ont.getId(), ont.getVersion());
	loadCategoryList (catList, ont);

        validate();
        // repaint();
//TEST  System.out.println("Exiting changeContext()");
	contextOnt = ont;
    }

    /** Submit the query to Parka and display the results in the resWin
     */
    public ResultSet doQuery()
    {
        // String        qstr;           // specifies query
        Query qry;
        ResultSet  rset;

        /* added for timing statistics */
        Date startTime, endTime;
        startTime = new Date();

        setCursor (Frame.WAIT_CURSOR);
        // qstr = qwin.getQueryString();
	qry = qwin.getQuery();
        // System.out.println ("Query = " + qry.toString());

	rset = kb.issueQuery(qry);

	// Remove duplicate entries, since they are unnecessary...
	rset.removeDuplicates();

/* TEST
        System.out.print ("Query Button:  Got ");
        if (rset == null)
            System.out.println ("0 answers");
        else
            System.out.println (rset.getNumResults() + " answers");
	System.out.println("rset = " + rset.toString());
*/
        updateQueryResults(rset);

        endTime = new Date();
        long duration = endTime.getTime() - startTime.getTime();
        // System.out.println("User time: " + duration + " ms");
        // System.out.println();

        setCursor (Frame.DEFAULT_CURSOR);

        return rset;
    }

    /** Notifies the results window of a new result set, and makes the
      calls to display them in the table and change the headings to match.
    */
    public void updateQueryResults(ResultSet rset)
    {
        String        headings[];

        if (rset != null)
	  {
            // convert the variable names into English descriptions
            headings = qwin.getVariableDescs(rset.getColVars());
            resWin.changeTable (headings, rset);
	    if (rset.getNumResults() == 1)
	      showStatus("1 answer.");
	    else
	      showStatus(rset.getNumResults() + " answers.");
        }
        else
        {
            resWin.changeTable (null, rset);
	    showStatus("Internal Error: rset is null");
            // resWin.newEmptyMultiColumn();
        }
    }

    /** Submit a Find query to Parka, merely to find number of answers
       available.  Will not fill up java data structures.
     */
    public int doFindQueryForNumAnswers (int index)
    {
      // String        qstr;           // specifies query
	Query         query;
        ResultSet  rset;

        // qstr = qwin.getFindQueryString (index);
	query = qwin.getFindQuery(index);
/* TEST
        System.out.println ("Find query is:\n" + qstr);
*/
	/* since we only want to know if there are any answers, there is
	   no need to retrieve more than one of them. */
	query.setNumAnswersRequested(1);

	rset = kb.issueQuery(query);

/* TEST
        System.out.print ("Query Button:  Got ");
        if (rset == null)
	    System.out.println ("0 answers");
	else
	    System.out.println (rset.getNumResults() + " answers");
*/
        if (rset == null)
	    return  0;
	else
	    return  (rset.getNumResults());
    }

    void doFindQuery (int index)
    {
	Query         query;
        ResultSet  rset;

        // String  findQuery = qwin.getFindNamesQueryString (index);
	query = qwin.getFindNamesQuery(index);

/* TEST
        System.out.println ("Find query is:\n" + findQuery);
*/
	rset = kb.issueQuery(query);

	// Remove duplicate entries, since they are unnecessary...
	rset.removeDuplicates();

        updateQueryResults(rset);
    }

    /** Implements ParkaWin interface function to display messages from the
       Parka server
    */
  /*
    public void displayMessage (String msg)
    {
        // System.out.println(msg);
    }
    */

  /** Returns a ShoeKb object which allows standard SHOE operations to be
    excuted against an underlying knowledge base. This function can be
    overriden so that a subclass of ShoeKb with an alternate implementation
    can be returned. */
  public ShoeKb getShoeKb(KBInterface kb, OntManager ontman,
			  PrintStream logOut) {

    return new ShoeKb(kb, ontman, logOut);
  }

  /** Returns a KB interface suitable for querying. This
      function can be overriden so that different backends can be
      used. */
  //this method is made abstract by ZPan on 5-30-02.
  //It will be implemented by ShoeParka.SS_Frame
    public abstract KBInterface getKbInterface();
  /*{
    //System.out.println("Using Parka for KB");
    //return new ParkaInterface();
    }*/

  /** Override resize so we can see preferred sizes for components. */
  /*
  public void reshape(int x, int y, int width, int height) {

    super.reshape(x, y, width, height);
    System.out.println("Pref sqwin = " + sqwin.getPreferredSize());
    System.out.println("Pref sqwinPanel = " + sqwinPanel.getPreferredSize());
  }
  */



  /* ******** Private Methods ********* */

  /** Loads the ontology proxy information from a local file. Each
    line specifies an ontology id,
    version and a proxy URL for retrieving it. These proxies are then
    set in the OntManager.
    */
  private void loadOntProxyForApplication() {

    FileInputStream inStream;
    String errMsg;

    File proxyFile = new File(PROXY_FILE);
    try {
      inStream = new FileInputStream(proxyFile);
      errMsg = ontman.loadOntProxyInputStream(inStream);
      if (errMsg != null) {
	System.out.println("Found errors while reading ontology proxy file " +
		   PROXY_FILE + ":\n" + errMsg);
	showStatus("Found errors while reading ontology proxy file " +
		   PROXY_FILE);
      }
      System.out.println("Loaded ont proxy file for application");
    } catch (FileNotFoundException e) {
      showStatus("There is no ontology proxy file: " + PROXY_FILE);
    }
  }

  /** Loads the ontology proxy file from the machine that hosts the
    applet. Each line specifies an ontology id,
    version and a proxy URL for retrieving it. These proxies are then
    set in the OntManager.
    */
  private void loadOntProxyForApplet() {

    URL         baseUrl;
    URL         proxyUrl = null;
    InputStream inStream;
    String      errMsg;

    if (srcApplet == null)
      System.out.println("Error in loadOntProxyForApplet(): " +
			 "No applet defined.");
    else {
      baseUrl = srcApplet.getCodeBase();
      try {
	proxyUrl = new URL(baseUrl, PROXY_FILE);
	inStream = proxyUrl.openStream();
	errMsg = ontman.loadOntProxyInputStream(inStream);
	if (errMsg != null) {
	  System.out.println("Found errors while reading ontology proxy file "+
			     proxyUrl.toString() + ":\n" + errMsg);
	  showStatus("Found errors while reading ontology proxy file " +
		     proxyUrl.toString());
	}
      } catch (MalformedURLException e) {
	showStatus("Could not create URL for proxy file");
      } catch (IOException e) {
	showStatus("There is no ontology proxy file: " + proxyUrl.toString());
      }
    }
  }

}


